from mpi4py import MPI
import mpiunittest as unittest
import sys, os, mpi4py

MPI4PYPATH = os.path.abspath(os.path.dirname(mpi4py.__path__[0]))

CHILDSCRIPT = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'spawn_child.py')
    )

HAVE_MPE = 'MPE_LOGFILE_PREFIX' in os.environ
HAVE_VT  = 'VT_FILE_PREFIX' in os.environ

def childscript():
    from tempfile import mkstemp
    from textwrap import dedent
    fd, script = mkstemp(suffix='.py', prefix="mpi4py-")
    os.close(fd)
    with open(script, "w") as f:
        f.write(dedent("""\
        #!%(python)s
        import sys; sys.path.insert(0, "%(path)s")
        import mpi4py
        if %(mpe)s: mpi4py.profile('mpe', logfile="%(logfile)s")
        if %(vt)s:  mpi4py.profile('vt',  logfile="%(logfile)s")
        from mpi4py import MPI
        parent = MPI.Comm.Get_parent()
        parent.Barrier()
        parent.Disconnect()
        assert parent == MPI.COMM_NULL
        parent = MPI.Comm.Get_parent()
        assert parent == MPI.COMM_NULL
        """ % dict(python=sys.executable, path=MPI4PYPATH,
                   mpe=HAVE_MPE, vt=HAVE_VT,
                   logfile="runtests-mpi4py-child")))
    os.chmod(script, int("770", 8))
    return script

def ch4_ucx():
    return 'ch4:ucx' in MPI.Get_library_version()

def ch4_ofi():
    return 'ch4:ofi' in MPI.Get_library_version()

def appnum():
    if MPI.APPNUM == MPI.KEYVAL_INVALID: return None
    return MPI.COMM_WORLD.Get_attr(MPI.APPNUM)

def badport():
    if MPI.get_vendor()[0] != 'MPICH':
        return False
    try:
        port = MPI.Open_port()
        MPI.Close_port(port)
    except:
        port = ""
    return port == ""

def using_GPU():
    # Once a CUDA context is created, the process cannot be forked.
    # Note: This seems to be a partial fix. Even if we are running cpu-only
    # tests, if MPI is built with CUDA support we can still fail. Unfortunately
    # there is no runtime check for us to detect if it's the case...
    disabled_cupy = (sys.modules.get('cupy', -1) is None)
    disabled_numba = (sys.modules.get('numba', -1) is None)
    return False if (disabled_cupy and disabled_numba) else True


@unittest.skipMPI('MPI(<2.0)')
@unittest.skipMPI('openmpi(<3.0.0)')
@unittest.skipMPI('openmpi(==4.0.0)')
@unittest.skipMPI('openmpi(==4.0.1)', sys.platform=='darwin')
@unittest.skipMPI('openmpi(==4.0.2)', sys.platform=='darwin')
@unittest.skipMPI('mpich', appnum() is None)
@unittest.skipMPI('mpich', badport())
@unittest.skipMPI('msmpi(<8.1.0)')
@unittest.skipMPI('msmpi', appnum() is None)
@unittest.skipMPI('MVAPICH2')
@unittest.skipMPI('MPICH2')
@unittest.skipMPI('MPICH1')
@unittest.skipMPI('PlatformMPI')
@unittest.skipIf(using_GPU(), 'using CUDA')
class BaseTestSpawn(object):

    COMM = MPI.COMM_NULL
    COMMAND = sys.executable
    ARGS = [CHILDSCRIPT, MPI4PYPATH,
            "mpe" if HAVE_MPE else
            "vt"  if HAVE_VT  else ""]
    MAXPROCS = 1
    INFO = MPI.INFO_NULL
    ROOT = 0

    def testCommSpawn(self):
        self.COMM.Barrier()
        child = self.COMM.Spawn(self.COMMAND, self.ARGS, self.MAXPROCS,
                                info=self.INFO, root=self.ROOT)
        local_size = child.Get_size()
        remote_size = child.Get_remote_size()
        child.Barrier()
        child.Disconnect()
        self.COMM.Barrier()
        self.assertEqual(local_size, self.COMM.Get_size())
        self.assertEqual(remote_size, self.MAXPROCS)

    @unittest.skipMPI('msmpi')
    def testErrcodes(self):
        self.COMM.Barrier()
        errcodes = []
        child = self.COMM.Spawn(self.COMMAND, self.ARGS, self.MAXPROCS,
                                info=self.INFO, root=self.ROOT,
                                errcodes=errcodes)
        child.Barrier()
        child.Disconnect()
        self.COMM.Barrier()
        self.assertEqual(len(errcodes), self.MAXPROCS)
        for errcode in errcodes:
            self.assertEqual(errcode, MPI.SUCCESS)

    @unittest.skipMPI('msmpi')
    @unittest.skipMPI('mpich(==3.4.1)', ch4_ofi())
    def testArgsOnlyAtRoot(self):
        self.COMM.Barrier()
        if self.COMM.Get_rank() == self.ROOT:
            child = self.COMM.Spawn(self.COMMAND, self.ARGS, self.MAXPROCS,
                                    info=self.INFO, root=self.ROOT)
        else:
            child = self.COMM.Spawn(None, None, -1,
                                    info=MPI.INFO_NULL, root=self.ROOT)
        child.Barrier()
        child.Disconnect()
        self.COMM.Barrier()

    @unittest.skipIf(os.name != 'posix', 'posix')
    def testNoArgs(self):
        self.COMM.Barrier()
        script = None
        if self.COMM.Get_rank() == self.ROOT:
            script = childscript()
        self.COMM.Barrier()
        script = self.COMM.bcast(script, root=self.ROOT)
        child = self.COMM.Spawn(script, None, self.MAXPROCS,
                                info=self.INFO, root=self.ROOT)
        child.Barrier()
        child.Disconnect()
        self.COMM.Barrier()
        if self.COMM.Get_rank() == self.ROOT:
            os.remove(script)
        self.COMM.Barrier()

    def testCommSpawnMultiple(self):
        self.COMM.Barrier()
        count = 2 + (self.COMM.Get_size() == 0)
        COMMAND = [self.COMMAND] * count
        ARGS = [self.ARGS] * len(COMMAND)
        MAXPROCS = [self.MAXPROCS] * len(COMMAND)
        INFO = [self.INFO] * len(COMMAND)
        child = self.COMM.Spawn_multiple(
            COMMAND, ARGS, MAXPROCS,
            info=INFO, root=self.ROOT)
        local_size = child.Get_size()
        remote_size = child.Get_remote_size()
        child.Barrier()
        child.Disconnect()
        self.COMM.Barrier()
        self.assertEqual(local_size, self.COMM.Get_size())
        self.assertEqual(remote_size, sum(MAXPROCS))

    def testCommSpawnMultipleDefaults1(self):
        self.COMM.Barrier()
        count = 2 + (self.COMM.Get_size() == 0)
        COMMAND = [self.COMMAND] * count
        ARGS = [self.ARGS] * len(COMMAND)
        child = self.COMM.Spawn_multiple(COMMAND, ARGS)
        local_size = child.Get_size()
        remote_size = child.Get_remote_size()
        child.Barrier()
        child.Disconnect()
        self.COMM.Barrier()
        self.assertEqual(local_size, self.COMM.Get_size())
        self.assertEqual(remote_size, len(COMMAND))

    def testCommSpawnMultipleDefaults2(self):
        self.COMM.Barrier()
        count = 2 + (self.COMM.Get_size() == 0)
        COMMAND = [self.COMMAND] * count
        ARGS = [self.ARGS] * len(COMMAND)
        child = self.COMM.Spawn_multiple(COMMAND, ARGS, 1, MPI.INFO_NULL)
        local_size = child.Get_size()
        remote_size = child.Get_remote_size()
        child.Barrier()
        child.Disconnect()
        self.COMM.Barrier()
        self.assertEqual(local_size, self.COMM.Get_size())
        self.assertEqual(remote_size, len(COMMAND))

    @unittest.skipMPI('msmpi')
    def testErrcodesMultiple(self):
        self.COMM.Barrier()
        count = 2 + (self.COMM.Get_size() == 0)
        COMMAND = [self.COMMAND] * count
        ARGS = [self.ARGS]*len(COMMAND)
        MAXPROCS = list(range(1, len(COMMAND)+1))
        INFO = MPI.INFO_NULL
        errcodelist = []
        child = self.COMM.Spawn_multiple(
            COMMAND, ARGS, MAXPROCS,
            info=INFO, root=self.ROOT,
            errcodes=errcodelist)
        child.Barrier()
        child.Disconnect()
        self.COMM.Barrier()
        self.assertEqual(len(errcodelist), len(COMMAND))
        for i, errcodes in enumerate(errcodelist):
            self.assertEqual(len(errcodes), MAXPROCS[i])
            for errcode in errcodes:
                self.assertEqual(errcode, MPI.SUCCESS)

    @unittest.skipMPI('msmpi')
    def testArgsOnlyAtRootMultiple(self):
        self.COMM.Barrier()
        if self.COMM.Get_rank() == self.ROOT:
            count = 2 + (self.COMM.Get_size() == 0)
            COMMAND = [self.COMMAND] * count
            ARGS = [self.ARGS] * len(COMMAND)
            MAXPROCS = list(range(1, len(COMMAND)+1))
            INFO = [MPI.INFO_NULL] * len(COMMAND)
            child = self.COMM.Spawn_multiple(
                COMMAND, ARGS, MAXPROCS,
                info=INFO, root=self.ROOT)
        else:
            child = self.COMM.Spawn_multiple(
                None, None, -1,
                info=MPI.INFO_NULL, root=self.ROOT)
        child.Barrier()
        child.Disconnect()
        self.COMM.Barrier()

    @unittest.skipIf(os.name != 'posix', 'posix')
    def testNoArgsMultiple(self):
        self.COMM.Barrier()
        script = None
        if self.COMM.Get_rank() == self.ROOT:
            script = childscript()
        self.COMM.Barrier()
        script = self.COMM.bcast(script, root=self.ROOT)
        count = 2 + (self.COMM.Get_size() == 0)
        COMMAND = [script] * count
        MAXPROCS = list(range(1, len(COMMAND)+1))
        INFO = [self.INFO] * len(COMMAND)
        child = self.COMM.Spawn_multiple(COMMAND, None, MAXPROCS,
                                         info=INFO, root=self.ROOT)
        child.Barrier()
        child.Disconnect()
        self.COMM.Barrier()
        if self.COMM.Get_rank() == self.ROOT:
            os.remove(script)
        self.COMM.Barrier()

class TestSpawnSelf(BaseTestSpawn, unittest.TestCase):
    COMM = MPI.COMM_SELF

class TestSpawnWorld(BaseTestSpawn, unittest.TestCase):
    COMM = MPI.COMM_WORLD

class TestSpawnSelfMany(BaseTestSpawn, unittest.TestCase):
    COMM = MPI.COMM_SELF
    MAXPROCS = MPI.COMM_WORLD.Get_size()

class TestSpawnWorldMany(BaseTestSpawn, unittest.TestCase):
    COMM = MPI.COMM_WORLD
    MAXPROCS = MPI.COMM_WORLD.Get_size()


if __name__ == '__main__':
    unittest.main()



from mpi4py import MPI
import helloworld as hw

null = MPI.COMM_NULL
fnull = null.py2f()
hw.sayhello(fnull)

comm = MPI.COMM_WORLD
fcomm = comm.py2f()
hw.sayhello(fcomm)

try:
    hw.sayhello(list())
except:
    pass
else:
    assert 0, "exception not raised"



from mpi4py import MPI

blens = (1, 1)
disps = (0, MPI.DOUBLE.size)
types = (MPI.DOUBLE, MPI.CHAR)
type1 = MPI.Datatype.Create_struct(blens, disps, types)

B = (2, 1, 3)
D = (0, 16, 26)
T = (MPI.FLOAT, type1, MPI.CHAR)
dtype = MPI.Datatype.Create_struct(B, D, T)

type1.Free()
dtype.Free()



#!/usr/bin/env python
"""
Parallel PI computation using Dynamic Process Management (DPM)
within Python objects exposing memory buffers (requires NumPy).

usage:

  + parent/child model::

      $ mpiexec -n 1 python cpi-dpm.py [nchilds]

  + client/server model::

      $ [xterm -e] mpiexec -n <nprocs> python cpi-dpm.py server [-v] &
      $ [xterm -e] mpiexec -n 1 python cpi-dpm.py client [-v]
"""

import sys
from mpi4py import MPI
import numpy as N

def get_n():
    prompt = "Enter the number of intervals: (0 quits) "
    try:
        n = int(input(prompt))
        if n < 0: n = 0
    except:
        n = 0
    return n

def view(pi, np=None, wt=None):
    from math import pi as PI
    prn = sys.stdout.write
    if pi is not None:
        prn("computed pi is:  %.16f\n"  % pi)
        prn("absolute error:  %.16f\n" % abs(pi - PI))
    if np is not None:
        prn("computing units: %d processes\n" % np)
    if wt is not None:
        prn("wall clock time: %g seconds\n" % wt)
    sys.stdout.flush()

def comp_pi(n, comm, root=0):
    nprocs = comm.Get_size()
    myrank = comm.Get_rank()
    n = N.array(n, 'i')
    comm.Bcast([n, MPI.INT], root=root)
    if n == 0: return 0.0
    h = 1.0 / n;
    s = 0.0;
    for i in range(myrank, n, nprocs):
        x = h * (i + 0.5);
        s += 4.0 / (1.0 + x**2);
    mypi = s * h
    mypi = N.array(mypi, 'd')
    pi   = N.array(0, 'd')
    comm.Reduce([mypi, MPI.DOUBLE],
                [pi,   MPI.DOUBLE],
                root=root, op=MPI.SUM)
    return pi

def master(icomm):
    n = get_n()
    wt = MPI.Wtime()
    n =  N.array(n, 'i')
    icomm.Send([n, MPI.INT], dest=0)
    pi = N.array(0, 'd')
    icomm.Recv([pi, MPI.DOUBLE], source=0)
    wt = MPI.Wtime() - wt
    if n == 0: return
    np = icomm.Get_remote_size()
    view(pi, np, wt)

def worker(icomm):
    myrank = icomm.Get_rank()
    if myrank == 0:
        source = dest = 0
    else:
        source = dest = MPI.PROC_NULL
    n =  N.array(0, 'i')
    icomm.Recv([n, MPI.INT], source=source)
    pi = comp_pi(n, comm=MPI.COMM_WORLD, root=0)
    pi = N.array(pi, 'd')
    icomm.Send([pi, MPI.DOUBLE], dest=dest)


# Parent/Child

def main_parent(nprocs=1):
    assert nprocs > 0
    assert MPI.COMM_WORLD.Get_size() == 1
    icomm = MPI.COMM_WORLD.Spawn(command=sys.executable,
                                 args=[__file__, 'child'],
                                 maxprocs=nprocs)
    master(icomm)
    icomm.Disconnect()

def main_child():
    icomm = MPI.Comm.Get_parent()
    assert icomm != MPI.COMM_NULL
    worker(icomm)
    icomm.Disconnect()

# Client/Server

def main_server(COMM):
    nprocs = COMM.Get_size()
    myrank = COMM.Get_rank()

    service, port, info = None, None, MPI.INFO_NULL
    if myrank == 0:
        port = MPI.Open_port(info)
        log(COMM, "open port '%s'", port)
        service = 'cpi'
        MPI.Publish_name(service, port, info)
        log(COMM, "service '%s' published.", service)
    else:
        port = ''

    log(COMM, "waiting for client connection ...")
    icomm = COMM.Accept(port, info, root=0)
    log(COMM, "client connection accepted.")

    worker(icomm)

    log(COMM, "disconnecting from client ...")
    icomm.Disconnect()
    log(COMM, "client disconnected.")

    if myrank == 0:
        MPI.Unpublish_name(service, port, info)
        log(COMM, "service '%s' unpublished", port)
        MPI.Close_port(port)
        log(COMM, "closed  port '%s' ", port)


def main_client(COMM):
    assert COMM.Get_size() == 1

    service, info = 'cpi', MPI.INFO_NULL
    port = MPI.Lookup_name(service, info)
    log(COMM, "service '%s' found in port '%s'.", service, port)

    log(COMM, "connecting to server ...")
    icomm = COMM.Connect(port, info, root=0)
    log(COMM, "server connected.")

    master(icomm)

    log(COMM, "disconnecting from server ...")
    icomm.Disconnect()
    log(COMM, "server disconnected.")


def main():
    assert len(sys.argv) <= 2

    if 'server' in sys.argv:
        main_server(MPI.COMM_WORLD)
    elif 'client' in sys.argv:
        main_client(MPI.COMM_WORLD)
    elif 'child'  in sys.argv:
        main_child()
    else:
        try:    nchilds = int(sys.argv[1])
        except: nchilds = 2
        main_parent(nchilds)


VERBOSE = False

def log(COMM, fmt, *args):
    if not VERBOSE: return
    if COMM.rank != 0: return
    sys.stdout.write(fmt % args)
    sys.stdout.write('\n')
    sys.stdout.flush()

if __name__ == '__main__':
    if '-v' in sys.argv:
        VERBOSE = True
        sys.argv.remove('-v')
    main()



import sys
try:
    import threading
    HAVE_THREADING = True
except ImportError:
    import dummy_threading as threading
    HAVE_THREADING = False

VERBOSE = False
#VERBOSE = True

import mpi4py.rc
mpi4py.rc.thread_level = 'multiple'
from mpi4py import MPI
import mpiunittest as unittest

pypy3_lt_50 = (hasattr(sys, 'pypy_version_info') and
               sys.version_info[0] == 3 and
               sys.pypy_version_info < (5, 0))

class TestMPIThreads(unittest.TestCase):

    def testThreadLevels(self):
        levels = [MPI.THREAD_SINGLE,
                  MPI.THREAD_FUNNELED,
                  MPI.THREAD_SERIALIZED,
                  MPI.THREAD_MULTIPLE]
        for i in range(len(levels)-1):
            self.assertTrue(levels[i] < levels[i+1])
        try:
            provided = MPI.Query_thread()
            self.assertTrue(provided in levels)
        except NotImplementedError:
            self.skipTest('mpi-query_thread')

    def testIsThreadMain(self):
        try:
            flag = MPI.Is_thread_main()
        except NotImplementedError:
            self.skipTest('mpi-is_thread_main')
        name = threading.current_thread().name
        main = (name == 'MainThread') or not HAVE_THREADING
        self.assertEqual(flag, main)
        if VERBOSE:
            log = lambda m: sys.stderr.write(m+'\n')
            log("%s: MPI.Is_thread_main() -> %s" % (name, flag))

    @unittest.skipIf(pypy3_lt_50, 'pypy3(<5.0)')
    def testIsThreadMainInThread(self):
        try:
            provided = MPI.Query_thread()
        except NotImplementedError:
            self.skipTest('mpi-query_thread')
        self.testIsThreadMain()
        T = [threading.Thread(target=self.testIsThreadMain) for _ in range(5)]
        if provided == MPI.THREAD_MULTIPLE:
            for t in T:
                t.start()
            for t in T:
                t.join()
        elif provided == MPI.THREAD_SERIALIZED:
            for t in T:
                t.start()
                t.join()
        else:
            self.skipTest('mpi-thread_level')


if __name__ == '__main__':
    unittest.main()



cdef class Errhandler:

    """
    Error handler
    """

    def __cinit__(self, Errhandler errhandler: Optional[Errhandler] = None):
        self.ob_mpi = MPI_ERRHANDLER_NULL
        if errhandler is None: return
        self.ob_mpi = errhandler.ob_mpi

    def __dealloc__(self):
        if not (self.flags & PyMPI_OWNED): return
        CHKERR( del_Errhandler(&self.ob_mpi) )

    def __richcmp__(self, other, int op):
        if not isinstance(other, Errhandler): return NotImplemented
        cdef Errhandler s = <Errhandler>self, o = <Errhandler>other
        if   op == Py_EQ: return (s.ob_mpi == o.ob_mpi)
        elif op == Py_NE: return (s.ob_mpi != o.ob_mpi)
        cdef mod = type(self).__module__
        cdef cls = type(self).__name__
        raise TypeError("unorderable type: '%s.%s'" % (mod, cls))

    def __bool__(self) -> bool:
        return self.ob_mpi != MPI_ERRHANDLER_NULL

    def Free(self) -> None:
        """
        Free an error handler
        """
        CHKERR( MPI_Errhandler_free(&self.ob_mpi) )
        if self is __ERRORS_RETURN__:    self.ob_mpi = MPI_ERRORS_RETURN
        if self is __ERRORS_ARE_FATAL__: self.ob_mpi = MPI_ERRORS_ARE_FATAL

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Errhandler_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Errhandler:
        """
        """
        cdef Errhandler errhandler = Errhandler.__new__(Errhandler)
        errhandler.ob_mpi = MPI_Errhandler_f2c(arg)
        return errhandler



cdef Errhandler __ERRHANDLER_NULL__  = new_Errhandler(MPI_ERRHANDLER_NULL)
cdef Errhandler __ERRORS_RETURN__    = new_Errhandler(MPI_ERRORS_RETURN)
cdef Errhandler __ERRORS_ARE_FATAL__ = new_Errhandler(MPI_ERRORS_ARE_FATAL)


# Predefined errhandler handles
# -----------------------------

ERRHANDLER_NULL  = __ERRHANDLER_NULL__  #: Null error handler
ERRORS_RETURN    = __ERRORS_RETURN__    #: Errors return error handler
ERRORS_ARE_FATAL = __ERRORS_ARE_FATAL__ #: Errors are fatal error handler



# --------------------------------------------------------------------

from mpi4py import MPI
from array import array as _array
import struct as _struct

class Counter(object):

    def __init__(self, comm):
        #
        size = comm.Get_size()
        rank = comm.Get_rank()
        #
        itemsize = MPI.INT.Get_size()
        if rank == 0:
            mem = MPI.Alloc_mem(itemsize*size, MPI.INFO_NULL)
            mem[:] = _struct.pack('i', 0) * size
        else:
            mem = MPI.BOTTOM
        self.win = MPI.Win.Create(mem, itemsize, MPI.INFO_NULL, comm)
        #
        blens = [rank, size-rank-1]
        disps = [0, rank+1]
        self.dt_get = MPI.INT.Create_indexed(blens, disps).Commit()
        #
        self.myval = 0

    def free(self):
        self.dt_get.Free()
        mem = self.win.tomemory()
        self.win.Free()
        if mem: MPI.Free_mem(mem)

    def next(self):
        #
        group  = self.win.Get_group()
        size = group.Get_size()
        rank = group.Get_rank()
        group.Free()
        #
        incr = _array('i', [1])
        vals = _array('i', [0])*size
        self.win.Lock(0)
        self.win.Accumulate([incr, 1, MPI.INT], 0,
                            [rank, 1, MPI.INT], MPI.SUM)
        self.win.Get([vals, 1, self.dt_get], 0,
                     [   0, 1, self.dt_get])
        self.win.Unlock(0)
        #
        vals[rank] = self.myval
        self.myval += 1
        nxtval = sum(vals)
        #
        return nxtval

# --------------------------------------------------------------------

def test():
    vals = []
    counter = Counter(MPI.COMM_WORLD)
    for i in range(5):
        c = counter.next()
        vals.append(c)
    counter.free()

    vals = MPI.COMM_WORLD.allreduce(vals)
    assert sorted(vals) == list(range(len(vals)))

if __name__ == '__main__':
    test()

# --------------------------------------------------------------------



from mpi4py import MPI
import mpiunittest as unittest


class BaseTestMessageZero(object):

    null_b = [None, MPI.INT]
    null_v = [None, (0, None), MPI.INT]

    def testPointToPoint(self):
        comm = self.COMM
        comm.Sendrecv(sendbuf=self.null_b,   dest=comm.rank,
                      recvbuf=self.null_b, source=comm.rank)
        r2 = comm.Irecv(self.null_b, comm.rank)
        r1 = comm.Isend(self.null_b, comm.rank)
        MPI.Request.Waitall([r1, r2])

    def testCollectivesBlock(self):
        comm = self.COMM
        comm.Bcast(self.null_b)
        comm.Gather(self.null_b, self.null_b)
        comm.Scatter(self.null_b, self.null_b)
        comm.Allgather(self.null_b, self.null_b)
        comm.Alltoall(self.null_b, self.null_b)

    def testCollectivesVector(self):
        comm = self.COMM
        comm.Gatherv(self.null_b, self.null_v)
        comm.Scatterv(self.null_v, self.null_b)
        comm.Allgatherv(self.null_b, self.null_v)
        comm.Alltoallv(self.null_v, self.null_v)

    @unittest.skipMPI('openmpi')
    @unittest.skipMPI('SpectrumMPI')
    def testReductions(self):
        comm = self.COMM
        comm.Reduce(self.null_b, self.null_b)
        comm.Allreduce(self.null_b, self.null_b)
        comm.Reduce_scatter_block(self.null_b, self.null_b)
        rcnt = [0]*comm.Get_size()
        comm.Reduce_scatter(self.null_b, self.null_b, rcnt)
        try: comm.Scan(self.null_b, self.null_b)
        except NotImplementedError: pass
        try: comm.Exscan(self.null_b, self.null_b)
        except NotImplementedError: pass

class TestMessageZeroSelf(BaseTestMessageZero, unittest.TestCase):
    COMM = MPI.COMM_SELF

class TestMessageZeroWorld(BaseTestMessageZero, unittest.TestCase):
    COMM = MPI.COMM_WORLD


if __name__ == '__main__':
    unittest.main()



from mpi4py import MPI
import mpiunittest as unittest
import arrayimpl


class BaseTestP2PBuf(object):

    COMM = MPI.COMM_NULL

    def testSendrecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        for array, typecode in arrayimpl.subTest(self):
            for s in range(0, size):
                sbuf = array( s, typecode, s)
                rbuf = array(-1, typecode, s+1)
                self.COMM.Sendrecv(sbuf.as_mpi(), dest,   0,
                                   rbuf.as_mpi(), source, 0)
                for value in rbuf[:-1]:
                    self.assertEqual(value, s)
                self.assertEqual(rbuf[-1], -1)

    def testSendrecvReplace(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        for array, typecode in arrayimpl.subTest(self):
            for s in range(0, size):
                buf = array(rank, typecode, s);
                self.COMM.Sendrecv_replace(buf.as_mpi(), dest, 0, source, 0)
                for value in buf:
                    self.assertEqual(value, source)

    def testSendRecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for array, typecode in arrayimpl.subTest(self):
            if unittest.is_mpi_gpu('openmpi', array): continue
            if unittest.is_mpi_gpu('mvapich2', array): continue
            for s in range(0, size):
                #
                sbuf = array( s, typecode, s)
                rbuf = array(-1, typecode, s)
                mem  = array( 0, typecode, 2*(s+MPI.BSEND_OVERHEAD)).as_raw()
                if size == 1:
                    MPI.Attach_buffer(mem)
                    rbuf = sbuf
                    MPI.Detach_buffer()
                elif rank == 0:
                    MPI.Attach_buffer(mem)
                    self.COMM.Ibsend(sbuf.as_mpi(), 1, 0).Wait()
                    self.COMM.Bsend(sbuf.as_mpi(), 1, 0)
                    MPI.Detach_buffer()
                    self.COMM.Send(sbuf.as_mpi(), 1, 0)
                    self.COMM.Ssend(sbuf.as_mpi(), 1, 0)
                    self.COMM.Recv(rbuf.as_mpi(),  1, 0)
                    self.COMM.Recv(rbuf.as_mpi(),  1, 0)
                    self.COMM.Recv(rbuf.as_mpi(), 1, 0)
                    self.COMM.Recv(rbuf.as_mpi(), 1, 0)
                elif rank == 1:
                    self.COMM.Recv(rbuf.as_mpi(), 0, 0)
                    self.COMM.Recv(rbuf.as_mpi(), 0, 0)
                    self.COMM.Recv(rbuf.as_mpi(), 0, 0)
                    self.COMM.Recv(rbuf.as_mpi(), 0, 0)
                    MPI.Attach_buffer(mem)
                    self.COMM.Ibsend(sbuf.as_mpi(), 0, 0).Wait()
                    self.COMM.Bsend(sbuf.as_mpi(), 0, 0)
                    MPI.Detach_buffer()
                    self.COMM.Send(sbuf.as_mpi(), 0, 0)
                    self.COMM.Ssend(sbuf.as_mpi(), 0, 0)
                else:
                    rbuf = sbuf
                for value in rbuf:
                    self.assertEqual(value, s)
                #
                rank = self.COMM.Get_rank()
                sbuf = array( s, typecode, s)
                rbuf = array(-1, typecode, s)
                rreq = self.COMM.Irecv(rbuf.as_mpi(), rank, 0)
                self.COMM.Rsend(sbuf.as_mpi(), rank, 0)
                rreq.Wait()
                for value in rbuf:
                    self.assertEqual(value, s)
                rbuf = array(-1, typecode, s)
                rreq = self.COMM.Irecv(rbuf.as_mpi(), rank, 0)
                self.COMM.Irsend(sbuf.as_mpi(), rank, 0).Wait()
                rreq.Wait()
                for value in rbuf:
                    self.assertEqual(value, s)

    def testProcNull(self):
        comm = self.COMM
        #
        comm.Sendrecv(None, MPI.PROC_NULL, 0,
                      None, MPI.PROC_NULL, 0)
        comm.Sendrecv_replace(None,
                              MPI.PROC_NULL, 0,
                              MPI.PROC_NULL, 0)
        #
        comm.Send (None, MPI.PROC_NULL)
        comm.Isend (None, MPI.PROC_NULL).Wait()
        #
        comm.Ssend(None, MPI.PROC_NULL)
        comm.Issend(None, MPI.PROC_NULL).Wait()
        #
        buf = MPI.Alloc_mem(MPI.BSEND_OVERHEAD)
        MPI.Attach_buffer(buf)
        comm.Bsend(None, MPI.PROC_NULL)
        comm.Ibsend(None, MPI.PROC_NULL).Wait()
        MPI.Detach_buffer()
        MPI.Free_mem(buf)
        #
        comm.Rsend(None, MPI.PROC_NULL)
        comm.Irsend(None, MPI.PROC_NULL).Wait()
        #
        comm.Recv (None, MPI.PROC_NULL)
        comm.Irecv(None, MPI.PROC_NULL).Wait()

    @unittest.skipMPI('mpich(==3.4.1)')
    def testProcNullPersistent(self):
        comm = self.COMM
        #
        req = comm.Send_init(None, MPI.PROC_NULL)
        req.Start(); req.Wait(); req.Free()
        #
        req = comm.Ssend_init(None, MPI.PROC_NULL)
        req.Start(); req.Wait(); req.Free()
        #
        buf = MPI.Alloc_mem(MPI.BSEND_OVERHEAD)
        MPI.Attach_buffer(buf)
        req = comm.Bsend_init(None, MPI.PROC_NULL)
        req.Start(); req.Wait(); req.Free()
        MPI.Detach_buffer()
        MPI.Free_mem(buf)
        #
        req = comm.Rsend_init(None, MPI.PROC_NULL)
        req.Start(); req.Wait(); req.Free()
        #
        req = comm.Recv_init(None, MPI.PROC_NULL)
        req.Start(); req.Wait(); req.Free()

    def testPersistent(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        for array, typecode in arrayimpl.subTest(self):
            if unittest.is_mpi_gpu('openmpi', array): continue
            if unittest.is_mpi_gpu('mvapich2', array): continue
            for s in range(size):
                for xs in range(3):
                    #
                    sbuf = array( s, typecode, s)
                    rbuf = array(-1, typecode, s+xs)
                    sendreq = self.COMM.Send_init(sbuf.as_mpi(), dest, 0)
                    recvreq = self.COMM.Recv_init(rbuf.as_mpi(), source, 0)
                    sendreq.Start()
                    recvreq.Start()
                    sendreq.Wait()
                    recvreq.Wait()
                    self.assertNotEqual(sendreq, MPI.REQUEST_NULL)
                    self.assertNotEqual(recvreq, MPI.REQUEST_NULL)
                    sendreq.Free()
                    recvreq.Free()
                    self.assertEqual(sendreq, MPI.REQUEST_NULL)
                    self.assertEqual(recvreq, MPI.REQUEST_NULL)
                    for value in rbuf[:s]:
                        self.assertEqual(value, s)
                    for value in rbuf[s:]:
                        self.assertEqual(value, -1)
                    #
                    sbuf = array(s,  typecode, s)
                    rbuf = array(-1, typecode, s+xs)
                    sendreq = self.COMM.Send_init(sbuf.as_mpi(), dest, 0)
                    recvreq = self.COMM.Recv_init(rbuf.as_mpi(), source, 0)
                    reqlist = [sendreq, recvreq]
                    MPI.Prequest.Startall(reqlist)
                    index1 = MPI.Prequest.Waitany(reqlist)
                    self.assertTrue(index1 in [0, 1])
                    self.assertNotEqual(reqlist[index1], MPI.REQUEST_NULL)
                    index2 = MPI.Prequest.Waitany(reqlist)
                    self.assertTrue(index2 in [0, 1])
                    self.assertNotEqual(reqlist[index2], MPI.REQUEST_NULL)
                    self.assertTrue(index1 != index2)
                    index3 = MPI.Prequest.Waitany(reqlist)
                    self.assertEqual(index3, MPI.UNDEFINED)
                    for preq in reqlist:
                        self.assertNotEqual(preq, MPI.REQUEST_NULL)
                        preq.Free()
                        self.assertEqual(preq, MPI.REQUEST_NULL)
                    for value in rbuf[:s]:
                        self.assertEqual(value, s)
                    for value in rbuf[s:]:
                        self.assertEqual(value, -1)
                    #
                    sbuf = array( s, typecode, s)
                    rbuf = array(-1, typecode, s+xs)
                    sendreq = self.COMM.Ssend_init(sbuf.as_mpi(), dest, 0)
                    recvreq = self.COMM.Recv_init(rbuf.as_mpi(), source, 0)
                    sendreq.Start()
                    recvreq.Start()
                    sendreq.Wait()
                    recvreq.Wait()
                    self.assertNotEqual(sendreq, MPI.REQUEST_NULL)
                    self.assertNotEqual(recvreq, MPI.REQUEST_NULL)
                    sendreq.Free()
                    recvreq.Free()
                    self.assertEqual(sendreq, MPI.REQUEST_NULL)
                    self.assertEqual(recvreq, MPI.REQUEST_NULL)
                    for value in rbuf[:s]:
                        self.assertEqual(value, s)
                    for value in rbuf[s:]:
                        self.assertEqual(value, -1)
                    #
                    mem = array( 0, typecode, s+MPI.BSEND_OVERHEAD).as_raw()
                    sbuf = array( s, typecode, s)
                    rbuf = array(-1, typecode, s+xs)
                    MPI.Attach_buffer(mem)
                    sendreq = self.COMM.Bsend_init(sbuf.as_mpi(), dest, 0)
                    recvreq = self.COMM.Recv_init(rbuf.as_mpi(), source, 0)
                    sendreq.Start()
                    recvreq.Start()
                    sendreq.Wait()
                    recvreq.Wait()
                    MPI.Detach_buffer()
                    self.assertNotEqual(sendreq, MPI.REQUEST_NULL)
                    self.assertNotEqual(recvreq, MPI.REQUEST_NULL)
                    sendreq.Free()
                    recvreq.Free()
                    self.assertEqual(sendreq, MPI.REQUEST_NULL)
                    self.assertEqual(recvreq, MPI.REQUEST_NULL)
                    for value in rbuf[:s]:
                        self.assertEqual(value, s)
                    for value in rbuf[s:]:
                        self.assertEqual(value, -1)
                    #
                    rank = self.COMM.Get_rank()
                    sbuf = array( s, typecode, s)
                    rbuf = array(-1, typecode, s+xs)
                    recvreq = self.COMM.Recv_init (rbuf.as_mpi(), rank, 0)
                    sendreq = self.COMM.Rsend_init(sbuf.as_mpi(), rank, 0)
                    recvreq.Start()
                    sendreq.Start()
                    recvreq.Wait()
                    sendreq.Wait()
                    self.assertNotEqual(sendreq, MPI.REQUEST_NULL)
                    self.assertNotEqual(recvreq, MPI.REQUEST_NULL)
                    sendreq.Free()
                    recvreq.Free()
                    self.assertEqual(sendreq, MPI.REQUEST_NULL)
                    self.assertEqual(recvreq, MPI.REQUEST_NULL)
                    for value in rbuf[:s]:
                        self.assertEqual(value, s)
                    for value in rbuf[s:]:
                        self.assertEqual(value, -1)

    def testProbe(self):
        comm = self.COMM.Dup()
        try:
            request = comm.Issend([None, 0, MPI.BYTE], comm.rank, 123)
            self.assertTrue(request)
            status = MPI.Status()
            comm.Probe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertEqual(status.source, comm.rank)
            self.assertEqual(status.tag, 123)
            self.assertTrue(request)
            flag = request.Test()
            self.assertTrue(request)
            self.assertFalse(flag)
            comm.Recv([None, 0, MPI.BYTE], comm.rank, 123)
            self.assertTrue(request)
            flag = False
            while not flag:
                flag = request.Test()
            self.assertFalse(request)
            self.assertTrue(flag)
        finally:
            comm.Free()

    @unittest.skipMPI('MPICH1')
    @unittest.skipMPI('LAM/MPI')
    def testProbeCancel(self):
        comm = self.COMM.Dup()
        try:
            request = comm.Issend([None, 0, MPI.BYTE], comm.rank, 123)
            status = MPI.Status()
            comm.Probe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertEqual(status.source, comm.rank)
            self.assertEqual(status.tag, 123)
            request.Cancel()
            self.assertTrue(request)
            status = MPI.Status()
            request.Get_status(status)
            cancelled = status.Is_cancelled()
            if not cancelled:
                comm.Recv([None, 0, MPI.BYTE], comm.rank, 123)
                request.Wait()
            else:
                request.Free()
        finally:
            comm.Free()

    def testIProbe(self):
        comm = self.COMM.Dup()
        try:
            f = comm.Iprobe()
            self.assertFalse(f)
            f = comm.Iprobe(MPI.ANY_SOURCE)
            self.assertFalse(f)
            f = comm.Iprobe(MPI.ANY_SOURCE, MPI.ANY_TAG)
            self.assertFalse(f)
            status = MPI.Status()
            f = comm.Iprobe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertFalse(f)
            self.assertEqual(status.source, MPI.ANY_SOURCE)
            self.assertEqual(status.tag,    MPI.ANY_TAG)
            self.assertEqual(status.error,  MPI.SUCCESS)
        finally:
            comm.Free()


class TestP2PBufSelf(BaseTestP2PBuf, unittest.TestCase):
    COMM = MPI.COMM_SELF

class TestP2PBufWorld(BaseTestP2PBuf, unittest.TestCase):
    COMM = MPI.COMM_WORLD

class TestP2PBufSelfDup(TestP2PBufSelf):
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()
    def tearDown(self):
        self.COMM.Free()

@unittest.skipMPI('openmpi(<1.4.0)', MPI.Query_thread() > MPI.THREAD_SINGLE)
class TestP2PBufWorldDup(TestP2PBufWorld):
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()
    def tearDown(self):
        self.COMM.Free()


if __name__ == '__main__':
    unittest.main()



import numpy as np
import time

tic = time.time()

x1 = -2.0
x2 =  1.0
y1 = -1.0
y2 =  1.0

w = 150
h = 100
maxit = 127

def mandelbrot(x, y, maxit):
    c = x + y*1j
    z = 0 + 0j
    it = 0
    while abs(z) < 2 and it < maxit:
        z = z**2 + c
        it += 1
    return it

dx = (x2 - x1) / w
dy = (y2 - y1) / h

C = np.empty([h, w], dtype='i')
for k in np.arange(h):
    y = y1 + k * dy
    for j in np.arange(w):
        x = x1 + j * dx
        C[k, j] = mandelbrot(x, y, maxit)

M = C

toc = time.time()
print('wall clock time: %8.2f seconds' % (toc-tic))

# eye candy (requires matplotlib)
if 1:
    try:
        from matplotlib import pyplot as plt
        plt.imshow(M, aspect='equal')
        try:
            plt.nipy_spectral()
        except AttributeError:
            plt.spectral()
        plt.pause(2)
    except:
        pass



from mpi4py import MPI
import mpiunittest as unittest

class GReqCtx(object):

    source = 3
    tag    = 7
    completed = False

    cancel_called = False
    free_called = False

    def query(self, status):
        status.Set_source(self.source)
        status.Set_tag(self.tag)
    def free(self):
        self.free_called = True
    def cancel(self, completed):
        self.cancel_called = True
        if completed is not self.completed:
            raise MPI.Exception(MPI.ERR_PENDING)
        

@unittest.skipMPI('MPI(<2.0)')
@unittest.skipMPI('openmpi(==4.1.0)')
class TestGrequest(unittest.TestCase):

    def testAll(self):
        ctx = GReqCtx()
        greq = MPI.Grequest.Start(ctx.query, ctx.free, ctx.cancel)
        self.assertFalse(greq.Test())
        self.assertFalse(ctx.free_called)
        greq.Cancel()
        self.assertTrue(ctx.cancel_called)
        ctx.cancel_called = False
        greq.Complete()
        ctx.completed = True
        greq.Cancel()
        self.assertTrue(ctx.cancel_called)
        status = MPI.Status()
        self.assertTrue(greq.Test(status))
        self.assertEqual(status.Get_source(), ctx.source)
        self.assertEqual(status.Get_tag(), ctx.tag)
        self.assertEqual(status.Get_error(), MPI.SUCCESS)
        greq.Wait()
        self.assertTrue(ctx.free_called)

    def testAll1(self):
        ctx = GReqCtx()
        greq = MPI.Grequest.Start(ctx.query, None, None)
        self.assertFalse(greq.Test())
        greq.Cancel()
        greq.Complete()
        status = MPI.Status()
        self.assertTrue(greq.Test(status))
        self.assertEqual(status.Get_source(), ctx.source)
        self.assertEqual(status.Get_tag(), ctx.tag)
        self.assertEqual(status.Get_error(), MPI.SUCCESS)
        self.assertFalse(status.Is_cancelled())
        greq.Wait()

    def testAll2(self):
        greq = MPI.Grequest.Start(None, None, None)
        self.assertFalse(greq.Test())
        greq.Cancel()
        greq.Complete()
        status = MPI.Status()
        self.assertTrue(greq.Test(status))
        self.assertEqual(status.Get_source(), MPI.ANY_SOURCE)
        self.assertEqual(status.Get_tag(), MPI.ANY_TAG)
        self.assertEqual(status.Get_error(), MPI.SUCCESS)
        self.assertFalse(status.Is_cancelled())
        greq.Wait()


if __name__ == '__main__':
    unittest.main()



from mpi4py import MPI
import mpiunittest as unittest


class TestMemory(unittest.TestCase):

    def testMemory1(self):
        for size in range(0, 10000, 100):
            try:
                mem1 = MPI.Alloc_mem(size)
                self.assertEqual(len(mem1), size)
                MPI.Free_mem(mem1)
            except NotImplementedError:
                self.skipTest('mpi-alloc_mem')

    def testMemory2(self):
        for size in range(0, 10000, 100):
            try:
                mem2 = MPI.Alloc_mem(size, MPI.INFO_NULL)
                self.assertEqual(len(mem2), size)
                MPI.Free_mem(mem2)
            except NotImplementedError:
                self.skipTest('mpi-alloc_mem')


if __name__ == '__main__':
    unittest.main()



from mpi4py import MPI
import mpiunittest as unittest
import sys


MPI_ERR_OP = MPI.ERR_OP

try:
    import array
except ImportError:
    array = None

def asarray(typecode, data):
    try:
        memoryview
        _tobytes = lambda s: memoryview(s).tobytes()
    except NameError:
        _tobytes = lambda s: buffer(s)[:]
    try:
        _frombytes = array.array.frombytes
    except AttributeError:
        _frombytes = array.array.fromstring
    a = array.array(typecode, [])
    _frombytes(a, _tobytes(data))
    return a

def mysum_obj(a, b):
    for i in range(len(a)):
        b[i] = a[i] + b[i]
    return b

def mysum_buf(a, b, dt):
    assert dt == MPI.INT
    assert len(a) == len(b)
    b[:] = mysum_obj(asarray('i', a), asarray('i', b))

def mysum(ba, bb, dt):
    if dt is None:
        return mysum_obj(ba, bb)
    else:
        return mysum_buf(ba, bb, dt)


class TestOp(unittest.TestCase):

    def testConstructor(self):
        op = MPI.Op()
        self.assertFalse(op)
        self.assertEqual(op, MPI.OP_NULL)

    @unittest.skipIf(array is None, 'array')
    def testCreate(self):
        for comm in [MPI.COMM_SELF, MPI.COMM_WORLD]:
            for commute in [True, False]:
                for N in range(4):
                    myop = MPI.Op.Create(mysum, commute)
                    self.assertFalse(myop.is_predefined)
                    if (hasattr(sys, 'pypy_version_info') and
                        comm.size > 1):
                        myop.Free()
                        continue
                    try:
                        # buffer(empty_array) returns
                        # the same non-NULL pointer !!!
                        if N == 0: continue
                        size = comm.Get_size()
                        rank = comm.Get_rank()
                        a = array.array('i', [i*(rank+1) for i in range(N)])
                        b = array.array('i', [0]*len(a))
                        comm.Allreduce([a, MPI.INT], [b, MPI.INT], myop)
                        scale = sum(range(1,size+1))
                        for i in range(N):
                            self.assertEqual(b[i], scale*i)
                        ret = myop(a, b)
                        self.assertTrue(ret is b)
                        for i in range(N):
                            self.assertEqual(b[i], a[i]+scale*i)
                        myop2 = MPI.Op(myop)
                        a = array.array('i', [1]*N)
                        b = array.array('i', [2]*N)
                        ret = myop2(a, b)
                        self.assertTrue(ret is b)
                        for i in range(N):
                            self.assertEqual(b[i], 3)
                        myop2 = None
                    finally:
                        myop.Free()

    def testCreateMany(self):
        N = 32 # max user-defined operations
        #
        ops = []
        for i in range(N):
            o = MPI.Op.Create(mysum)
            ops.append(o)
        for o in ops: o.Free() # cleanup
        # another round
        ops = []
        for i in range(N):
            o = MPI.Op.Create(mysum)
            ops.append(o)
        for o in ops: o.Free() # cleanup

    def _test_call(self, op, args, res):
        self.assertEqual(op(*args), res)

    def testCall(self):
        self._test_call(MPI.MIN,  (2,3), 2)
        self._test_call(MPI.MAX,  (2,3), 3)
        self._test_call(MPI.SUM,  (2,3), 5)
        self._test_call(MPI.PROD, (2,3), 6)
        def xor(x,y): return bool(x) ^ bool(y)
        for x, y in ((0, 0),
                     (0, 1),
                     (1, 0),
                     (1, 1)):
            self._test_call(MPI.LAND,  (x,y), x and y)
            self._test_call(MPI.LOR,   (x,y), x or  y)
            self._test_call(MPI.LXOR,  (x,y), xor(x, y))
            self._test_call(MPI.BAND,  (x,y), x  &  y)
            self._test_call(MPI.BOR,   (x,y), x  |  y)
            self._test_call(MPI.BXOR,  (x,y), x  ^  y)
        if MPI.REPLACE:
            self._test_call(MPI.REPLACE, (2,3), 3)
            self._test_call(MPI.REPLACE, (3,2), 2)
        if MPI.NO_OP:
            self._test_call(MPI.NO_OP, (2,3), 2)
            self._test_call(MPI.NO_OP, (3,2), 3)

    def testMinMax(self):
        x = [1]; y = [1]
        res = MPI.MIN(x, y)
        self.assertTrue(res is x)
        res = MPI.MAX(x, y)
        self.assertTrue(res is x)

    def testMinMaxLoc(self):
        x = [1]; i = [2]; u = [x, i]
        y = [2]; j = [1]; v = [y, j]
        res = MPI.MINLOC(u, v)
        self.assertTrue(res[0] is x)
        self.assertTrue(res[1] is i)
        res = MPI.MINLOC(v, u)
        self.assertTrue(res[0] is x)
        self.assertTrue(res[1] is i)
        res = MPI.MAXLOC(u, v)
        self.assertTrue(res[0] is y)
        self.assertTrue(res[1] is j)
        res = MPI.MAXLOC(v, u)
        self.assertTrue(res[0] is y)
        self.assertTrue(res[1] is j)
        #
        x = [1]; i = 0; u = [x, i]
        y = [1]; j = 1; v = [y, j]
        res = MPI.MINLOC(u, v)
        self.assertTrue(res[0] is x)
        self.assertTrue(res[1] is i)
        res = MPI.MAXLOC(u, v)
        self.assertTrue(res[0] is x)
        self.assertTrue(res[1] is i)
        #
        x = [1]; i = 1; u = [x, i]
        y = [1]; j = 0; v = [y, j]
        res = MPI.MINLOC(u, v)
        self.assertTrue(res[0] is y)
        self.assertTrue(res[1] is j)
        res = MPI.MAXLOC(u, v)
        self.assertTrue(res[0] is y)
        self.assertTrue(res[1] is j)
        #
        x = [1]; i = [0]; u = [x, i]
        y = [1]; j = [1]; v = [y, j]
        res = MPI.MINLOC(u, v)
        self.assertTrue(res[0] is x)
        self.assertTrue(res[1] is i)
        res = MPI.MAXLOC(u, v)
        self.assertTrue(res[0] is x)
        self.assertTrue(res[1] is i)
        #
        x = [1]; i = [1]; u = [x, i]
        y = [1]; j = [0]; v = [y, j]
        res = MPI.MINLOC(u, v)
        self.assertTrue(res[0] is y)
        self.assertTrue(res[1] is j)
        res = MPI.MAXLOC(u, v)
        self.assertTrue(res[0] is y)
        self.assertTrue(res[1] is j)

    @unittest.skipMPI('openmpi(<=1.8.1)')
    def testIsCommutative(self):
        try:
            MPI.SUM.Is_commutative()
        except NotImplementedError:
            self.skipTest('mpi-op-is_commutative')
        ops = [MPI.MAX,    MPI.MIN,
               MPI.SUM,    MPI.PROD,
               MPI.LAND,   MPI.BAND,
               MPI.LOR,    MPI.BOR,
               MPI.LXOR,   MPI.BXOR,
               MPI.MAXLOC, MPI.MINLOC,]
        for op in ops:
            flag = op.Is_commutative()
            self.assertEqual(flag, op.is_commutative)
            self.assertTrue(flag)

    @unittest.skipMPI('openmpi(<=1.8.1)')
    @unittest.skipMPI('mpich(==3.4.1)')
    def testIsCommutativeExtra(self):
        try:
            MPI.SUM.Is_commutative()
        except NotImplementedError:
            self.skipTest('mpi-op-is_commutative')
        ops =  [MPI.REPLACE, MPI.NO_OP]
        for op in ops:
            if not op: continue
            flag = op.Is_commutative()
            self.assertEqual(flag, op.is_commutative)
            #self.assertFalse(flag)

    def testIsPredefined(self):
        self.assertTrue(MPI.OP_NULL.is_predefined)
        ops = [MPI.MAX,    MPI.MIN,
               MPI.SUM,    MPI.PROD,
               MPI.LAND,   MPI.BAND,
               MPI.LOR,    MPI.BOR,
               MPI.LXOR,   MPI.BXOR,
               MPI.MAXLOC, MPI.MINLOC,]
        for op in ops:
            self.assertTrue(op.is_predefined)


if __name__ == '__main__':
    unittest.main()



from mpi4py import MPI
import mpiunittest as unittest


class TestErrhandler(unittest.TestCase):

    def testPredefined(self):
        self.assertFalse(MPI.ERRHANDLER_NULL)
        self.assertTrue(MPI.ERRORS_ARE_FATAL)
        self.assertTrue(MPI.ERRORS_RETURN)

    def testCommGetSetErrhandler(self):
        for COMM in [MPI.COMM_SELF, MPI.COMM_WORLD]:
            for ERRHANDLER in [MPI.ERRORS_ARE_FATAL, MPI.ERRORS_RETURN,
                               MPI.ERRORS_ARE_FATAL, MPI.ERRORS_RETURN, ]:
                errhdl_1 = COMM.Get_errhandler()
                self.assertNotEqual(errhdl_1, MPI.ERRHANDLER_NULL)
                COMM.Set_errhandler(ERRHANDLER)
                errhdl_2 = COMM.Get_errhandler()
                self.assertEqual(errhdl_2, ERRHANDLER)
                errhdl_2.Free()
                self.assertEqual(errhdl_2, MPI.ERRHANDLER_NULL)
                COMM.Set_errhandler(errhdl_1)
                errhdl_1.Free()
                self.assertEqual(errhdl_1, MPI.ERRHANDLER_NULL)

    def testGetErrhandler(self):
        errhdls = []
        for i in range(100):
            e = MPI.COMM_WORLD.Get_errhandler()
            errhdls.append(e)
        for e in errhdls:
            e.Free()
        for e in errhdls:
            self.assertEqual(e, MPI.ERRHANDLER_NULL)

    @unittest.skipMPI('MPI(<2.0)')
    def testCommCallErrhandler(self):
        errhdl = MPI.COMM_SELF.Get_errhandler()
        comm = MPI.COMM_SELF.Dup()
        comm.Set_errhandler(MPI.ERRORS_RETURN)
        comm.Call_errhandler(MPI.ERR_OTHER)
        comm.Free()

    @unittest.skipMPI('MPI(<2.0)')
    @unittest.skipMPI('SpectrumMPI')
    def testWinCallErrhandler(self):
        try:
            win = MPI.Win.Create(MPI.BOTTOM, 1, MPI.INFO_NULL, MPI.COMM_SELF)
        except NotImplementedError:
            self.skipTest('mpi-win')
        win.Set_errhandler(MPI.ERRORS_RETURN)
        win.Call_errhandler(MPI.ERR_OTHER)
        win.Free()

    @unittest.skipMPI('MPI(<2.0)')
    @unittest.skipMPI('msmpi')
    def testFileCallErrhandler(self):
        import os, tempfile
        rank = MPI.COMM_WORLD.Get_rank()
        fd, filename = tempfile.mkstemp(prefix='mpi4py-', suffix="-%d"%rank)
        os.close(fd)
        amode = MPI.MODE_WRONLY | MPI.MODE_CREATE | MPI.MODE_DELETE_ON_CLOSE
        try:
            file = MPI.File.Open(MPI.COMM_SELF, filename, amode, MPI.INFO_NULL)
        except NotImplementedError:
            self.skipTest('mpi-file')
        file.Set_errhandler(MPI.ERRORS_RETURN)
        #file.Call_errhandler(MPI.ERR_OTHER)
        file.Call_errhandler(MPI.SUCCESS)
        file.Close()


try:
    MPI.Win.Create(MPI.BOTTOM, 1, MPI.INFO_NULL, MPI.COMM_SELF).Free()
except (NotImplementedError, MPI.Exception):
    TestErrhandler.testWinCallErrhandler = \
    unittest.disable(TestErrhandler.testWinCallErrhandler, 'mpi-win')


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------

from mpi4py import MPI
import sys, os

class Counter(object):

    def __init__(self, comm):
        assert not comm.Is_inter()
        self.comm = comm.Dup()
        # start counter process
        script  = os.path.abspath(__file__)
        if script[-4:] in ('.pyc', '.pyo'):
            script = script[:-1]
        self.child = self.comm.Spawn(sys.executable,
                                     [script, '--child'], 1)

    def free(self):
        self.comm.Barrier()
        # stop counter process
        rank = self.child.Get_rank()
        if rank == 0:
            self.child.send(None, 0, 1)
        self.child.Disconnect()
        #
        self.comm.Free()

    def next(self):
        #
        incr = 1
        self.child.send(incr, 0, 0)
        ival = self.child.recv(None, 0, 0)
        nxtval = ival
        #
        return nxtval

# --------------------------------------------------------------------

def _counter_child():
    parent = MPI.Comm.Get_parent()
    assert parent != MPI.COMM_NULL
    try:
        counter = 0
        status = MPI.Status()
        any_src, any_tag = MPI.ANY_SOURCE, MPI.ANY_TAG
        while True: # server loop
            incr = parent.recv(None, any_src, any_tag, status)
            if status.tag == 1: break
            parent.send(counter, status.source, 0)
            counter += incr
    finally:
        parent.Disconnect()

if __name__ == '__main__':
    if (len(sys.argv) > 1 and
        sys.argv[0] == __file__ and
        sys.argv[1] == '--child'):
        _counter_child()
        sys.exit(0)

# --------------------------------------------------------------------

def test():
    vals = []
    counter = Counter(MPI.COMM_WORLD)
    for i in range(5):
        c = counter.next()
        vals.append(c)
    counter.free()
    #
    vals = MPI.COMM_WORLD.allreduce(vals)
    assert sorted(vals) == list(range(len(vals)))

if __name__ == '__main__':
    test()

# --------------------------------------------------------------------



import sys, os, platform

from distutils.util  import split_quoted
from distutils.spawn import find_executable
from distutils import log as dulog

try:
    from collections import OrderedDict
except ImportError:
    OrderedDict = dict

try:
    from configparser import ConfigParser
    from configparser import Error as ConfigParserError
except ImportError:
    from ConfigParser import ConfigParser
    from ConfigParser import Error as ConfigParserError

class Config(object):

    def __init__(self, logger=None):
        self.log = logger or dulog
        self.section  = None
        self.filename = None
        self.compiler_info = OrderedDict((
                ('mpicc'  , None),
                ('mpicxx' , None),
                ('mpifort', None),
                ('mpif90' , None),
                ('mpif77' , None),
                ('mpild'  , None),
                ))
        self.library_info = OrderedDict((
            ('define_macros'        , []),
            ('undef_macros'         , []),
            ('include_dirs'         , []),

            ('libraries'            , []),
            ('library_dirs'         , []),
            ('runtime_library_dirs' , []),

            ('extra_compile_args'   , []),
            ('extra_link_args'      , []),
            ('extra_objects'        , []),
            ))

    def __bool__(self):
        for v in self.compiler_info.values():
            if v:
                return True
        for v in self.library_info.values():
            if v:
                return True
        return False

    __nonzero__ = __bool__

    def get(self, k, d=None):
        if k in self.compiler_info:
            return self.compiler_info[k]
        if k in self.library_info:
            return self.library_info[k]
        return d

    def info(self, log=None):
        if log is None: log = self.log
        mpicc   = self.compiler_info.get('mpicc')
        mpicxx  = self.compiler_info.get('mpicxx')
        mpifort = self.compiler_info.get('mpifort')
        mpif90  = self.compiler_info.get('mpif90')
        mpif77  = self.compiler_info.get('mpif77')
        mpild   = self.compiler_info.get('mpild')
        if mpicc:
            log.info("MPI C compiler:    %s", mpicc)
        if mpicxx:
            log.info("MPI C++ compiler:  %s", mpicxx)
        if mpifort:
            log.info("MPI F compiler:    %s", mpifort)
        if mpif90:
            log.info("MPI F90 compiler:  %s", mpif90)
        if mpif77:
            log.info("MPI F77 compiler:  %s", mpif77)
        if mpild:
            log.info("MPI linker:        %s", mpild)

    def update(self, config, **more):
        if hasattr(config, 'keys'):
            config = config.items()
        for option, value in config:
            if option in self.compiler_info:
                self.compiler_info[option] = value
            if option in self.library_info:
                self.library_info[option] = value
        if more:
            self.update(more)

    def setup(self, options, environ=None):
        if environ is None: environ = os.environ
        self.setup_library_info(options, environ)
        self.setup_compiler_info(options, environ)

    def setup_library_info(self, options, environ):
        filename = section = None
        mpiopt = getattr(options, 'mpi', None)
        mpiopt = environ.get('MPICFG', mpiopt)
        if mpiopt:
            if ',' in mpiopt:
                section, filename = mpiopt.split(',', 1)
            elif os.path.isfile(mpiopt):
                filename = mpiopt
            else:
                section = mpiopt
        if not filename: filename = "mpi.cfg"
        if not section:  section  = "mpi"

        mach = platform.machine()
        arch = platform.architecture()[0]
        plat = sys.platform
        osnm = os.name
        if   'linux' == plat[:5]: plat = 'linux'
        elif 'sunos' == plat[:5]: plat = 'solaris'
        elif 'win'   == plat[:3]: plat = 'windows'
        suffixes = []
        suffixes.append(plat+'-'+mach)
        suffixes.append(plat+'-'+arch)
        suffixes.append(plat)
        suffixes.append(osnm+'-'+mach)
        suffixes.append(osnm+'-'+arch)
        suffixes.append(osnm)
        suffixes.append(mach)
        suffixes.append(arch)
        sections  = [section+"-"+s for s in suffixes]
        sections += [section]
        self.load(filename, sections)
        if not self:
            if os.name == 'posix':
                self._setup_posix()
            if sys.platform == 'win32':
                self._setup_windows()

    def _setup_posix(self):
        pass

    def _setup_windows(self):
        if 'I_MPI_ROOT' in os.environ:
            self._setup_windows_intelmpi()
        else:
            self._setup_windows_msmpi()

    def _setup_windows_intelmpi(self):
        I_MPI_ROOT = os.environ.get('I_MPI_ROOT', '')
        if not I_MPI_ROOT: return
        if not os.path.isdir(I_MPI_ROOT): return
        arch = platform.architecture()[0][:2]
        archdir = {'32':'ia32', '64':'intel64'}[arch]
        mpi_dir = os.path.join(I_MPI_ROOT, archdir)
        if not os.path.isdir(mpi_dir): return
        IMPI_INC = os.path.join(mpi_dir, 'include')
        IMPI_LIB = os.path.join(mpi_dir, 'lib', 'release')
        if not os.path.isdir(IMPI_INC): return
        if not os.path.isdir(IMPI_LIB): return
        self.library_info.update(
            include_dirs=[IMPI_INC],
            library_dirs=[IMPI_LIB],
            libraries=['impi'])
        self.section = 'intelmpi'
        self.filename = [mpi_dir]
        return True

    def _setup_windows_msmpi(self):
        # Microsoft MPI (v7, v6, v5, v4)
        def msmpi_ver():
            try:
                try:
                    import winreg
                except ImportError:
                    import _winreg as winreg
                HKLM = winreg.HKEY_LOCAL_MACHINE
                subkey = r"SOFTWARE\Microsoft\MPI"
                with winreg.OpenKey(HKLM, subkey) as key:
                    for i in range(winreg.QueryInfoKey(key)[1]):
                        name, value, type = winreg.EnumValue(key, i)
                        if name != "Version": continue
                        major, minor = value.split('.')[:2]
                        return (int(major), int(minor))
            except: pass
            return (1, 0)
        def setup_msmpi(MSMPI_INC, MSMPI_LIB):
            from os.path import join, isfile
            ok = (MSMPI_INC and isfile(join(MSMPI_INC, 'mpi.h')) and
                  MSMPI_LIB and isfile(join(MSMPI_LIB, 'msmpi.lib')))
            if not ok: return False
            major, minor = msmpi_ver()
            MSMPI_VER = hex((major<<8)|(minor&0xFF))
            MSMPI_INC = os.path.normpath(MSMPI_INC)
            MSMPI_LIB = os.path.normpath(MSMPI_LIB)
            self.library_info.update(
                define_macros=[('MSMPI_VER', MSMPI_VER)],
                include_dirs=[MSMPI_INC],
                library_dirs=[MSMPI_LIB],
                libraries=['msmpi'])
            self.section = 'msmpi'
            self.filename = [os.path.dirname(MSMPI_INC)]
            return True
        arch = platform.architecture()[0][:2]
        # Look for Microsoft MPI in the environment
        MSMPI_INC = os.environ.get('MSMPI_INC')
        MSMPI_LIB = os.environ.get('MSMPI_LIB'+arch)
        if setup_msmpi(MSMPI_INC, MSMPI_LIB): return
        # Look for Microsoft MPI v7/v6/v5 in default install path
        for ProgramFiles in ('ProgramFiles', 'ProgramFiles(x86)'):
            ProgramFiles = os.environ.get(ProgramFiles, '')
            archdir = {'32':'x86', '64':'x64'}[arch]
            MSMPI_DIR = os.path.join(ProgramFiles, 'Microsoft SDKs', 'MPI')
            MSMPI_INC = os.path.join(MSMPI_DIR, 'Include')
            MSMPI_LIB = os.path.join(MSMPI_DIR, 'Lib', archdir)
            if setup_msmpi(MSMPI_INC, MSMPI_LIB): return
        # Look for Microsoft HPC Pack 2012 R2 in default install path
        for ProgramFiles in ('ProgramFiles', 'ProgramFiles(x86)'):
            ProgramFiles = os.environ.get(ProgramFiles, '')
            archdir = {'32':'i386', '64':'amd64'}[arch]
            MSMPI_DIR = os.path.join(ProgramFiles, 'Microsoft MPI')
            MSMPI_INC = os.path.join(MSMPI_DIR, 'Inc')
            MSMPI_LIB = os.path.join(MSMPI_DIR, 'Lib', archdir)
            if setup_msmpi(MSMPI_INC, MSMPI_LIB): return

        # Microsoft MPI (legacy) and others
        from glob import glob
        ProgramFiles = os.environ.get('ProgramFiles', '')
        CCP_HOME = os.environ.get('CCP_HOME', '')
        for (name, prefix, suffix) in (
            ('msmpi',    CCP_HOME,     ''),
            ('msmpi',    ProgramFiles, 'Microsoft HPC Pack 2012 R2'),
            ('msmpi',    ProgramFiles, 'Microsoft HPC Pack 2012'),
            ('msmpi',    ProgramFiles, 'Microsoft HPC Pack 2012 SDK'),
            ('msmpi',    ProgramFiles, 'Microsoft HPC Pack 2008 R2'),
            ('msmpi',    ProgramFiles, 'Microsoft HPC Pack 2008'),
            ('msmpi',    ProgramFiles, 'Microsoft HPC Pack 2008 SDK'),
            ):
            mpi_dir = os.path.join(prefix, suffix)
            if not mpi_dir or not os.path.isdir(mpi_dir): continue
            define_macros = []
            include_dir = os.path.join(mpi_dir, 'include')
            library = 'mpi'
            library_dir = os.path.join(mpi_dir, 'lib')
            if name == 'msmpi':
                include_dir = os.path.join(mpi_dir, 'inc')
                library = 'msmpi'
                arch = platform.architecture()[0]
                if arch == '32bit':
                    library_dir = os.path.join(library_dir, 'i386')
                if arch == '64bit':
                    library_dir = os.path.join(library_dir, 'amd64')
                if not os.path.isdir(include_dir):
                    include_dir = os.path.join(mpi_dir, 'include')
            self.library_info.update(
                define_macros=define_macros,
                include_dirs=[include_dir],
                libraries=[library],
                library_dirs=[library_dir],
                )
            self.section = name
            self.filename = [mpi_dir]
            break


    def setup_compiler_info(self, options, environ):
        def find_exe(cmd, path=None):
            if not cmd: return None
            parts = split_quoted(cmd)
            exe, args = parts[0], parts[1:]
            if not os.path.isabs(exe) and path:
                exe = os.path.basename(exe)
            exe = find_executable(exe, path)
            if not exe: return None
            return ' '.join([exe]+args)
        COMPILERS = (
            ('mpicc',   ['mpicc',   'mpcc_r']),
            ('mpicxx',  ['mpicxx',  'mpic++', 'mpiCC', 'mpCC_r']),
            ('mpifort', ['mpifort', 'mpif90', 'mpif77', 'mpfort_r']),
            ('mpif90',  ['mpif90',  'mpf90_r']),
            ('mpif77',  ['mpif77',  'mpf77_r']),
            ('mpild',   []),
            )
        #
        compiler_info = {}
        PATH = environ.get('PATH', '')
        for name, _ in COMPILERS:
            cmd = (environ.get(name.upper()) or
                   getattr(options, name, None) or
                   self.compiler_info.get(name) or
                   None)
            if cmd:
                exe = find_exe(cmd, path=PATH)
                if exe:
                    path = os.path.dirname(exe)
                    PATH = path + os.path.pathsep + PATH
                    compiler_info[name] = exe
                else:
                    self.log.error("error: '%s' not found", cmd)
        #
        if not self and not compiler_info:
            for name, candidates in COMPILERS:
                for cmd in candidates:
                    cmd = find_exe(cmd)
                    if cmd:
                        compiler_info[name] = cmd
                        break
        #
        self.compiler_info.update(compiler_info)


    def load(self, filename="mpi.cfg", section='mpi'):
        if isinstance(filename, str):
            filenames = filename.split(os.path.pathsep)
        else:
            filenames = list(filename)
        if isinstance(section, str):
            sections = section.split(',')
        else:
            sections = list(section)
        #
        try:
            parser = ConfigParser(dict_type=OrderedDict)
        except TypeError:
            parser = ConfigParser()
        try:
            read_ok = parser.read(filenames)
        except ConfigParserError:
            self.log.error(
                "error: parsing configuration file/s '%s'",
                os.path.pathsep.join(filenames))
            return None
        for section in sections:
            if parser.has_section(section):
                break
            section = None
        if not section:
            self.log.error(
                "error: section/s '%s' not found in file/s '%s'",
                ','.join(sections), os.path.pathsep.join(filenames))
            return None
        parser_items = list(parser.items(section, vars=None))
        #
        compiler_info = type(self.compiler_info)()
        for option, value in parser_items:
            if option in self.compiler_info:
                compiler_info[option] = value
        #
        pathsep = os.path.pathsep
        expanduser = os.path.expanduser
        expandvars = os.path.expandvars
        library_info = type(self.library_info)()
        for k, v in parser_items:
            if k in ('define_macros',
                     'undef_macros',
                     ):
                macros = [e.strip() for e in v.split(',')]
                if k == 'define_macros':
                    for i, m in enumerate(macros):
                        try: # -DFOO=bar
                            idx = m.index('=')
                            macro = (m[:idx], m[idx+1:] or None)
                        except ValueError: # -DFOO
                            macro = (m, None)
                        macros[i] = macro
                library_info[k] = macros
            elif k in ('include_dirs',
                       'library_dirs',
                       'runtime_dirs',
                       'runtime_library_dirs',
                       ):
                if k == 'runtime_dirs': k = 'runtime_library_dirs'
                pathlist = [p.strip() for p in v.split(pathsep)]
                library_info[k] = [expanduser(expandvars(p))
                                   for p in pathlist if p]
            elif k == 'libraries':
                library_info[k] = [e.strip() for e in split_quoted(v)]
            elif k in ('extra_compile_args',
                       'extra_link_args',
                       ):
                library_info[k] = split_quoted(v)
            elif k == 'extra_objects':
                library_info[k] = [expanduser(expandvars(e))
                                   for e in split_quoted(v)]
            elif hasattr(self, k):
                library_info[k] = v.strip()
            else:
                pass
        #
        self.section = section
        self.filename = read_ok
        self.compiler_info.update(compiler_info)
        self.library_info.update(library_info)
        return compiler_info, library_info, section, read_ok

    def dump(self, filename=None, section='mpi'):
        # prepare configuration values
        compiler_info   = self.compiler_info.copy()
        library_info = self.library_info.copy()
        for k in library_info:
            if k in ('define_macros',
                     'undef_macros',
                     ):
                macros = library_info[k]
                if k == 'define_macros':
                    for i, (m, v) in enumerate(macros):
                        if v is None:
                            macros[i] = m
                        else:
                            macros[i] = '%s=%s' % (m, v)
                library_info[k] = ','.join(macros)
            elif k in ('include_dirs',
                       'library_dirs',
                       'runtime_library_dirs',
                       ):
                library_info[k] = os.path.pathsep.join(library_info[k])
            elif isinstance(library_info[k], list):
                library_info[k] = ' '.join(library_info[k])
        # fill configuration parser
        try:
            parser = ConfigParser(dict_type=OrderedDict)
        except TypeError:
            parser = ConfigParser()
        parser.add_section(section)
        for option, value in compiler_info.items():
            if not value: continue
            parser.set(section, option, value)
        for option, value in library_info.items():
            if not value: continue
            parser.set(section, option, value)
        # save configuration file
        if filename is None:
            parser.write(sys.stdout)
        elif hasattr(filename, 'write'):
            parser.write(filename)
        elif isinstance(filename, str):
            with open(filename, 'w') as f:
                parser.write(f)
        return parser


if __name__ == '__main__':
    import optparse
    parser = optparse.OptionParser()
    parser.add_option("--mpi",     type="string")
    parser.add_option("--mpicc",   type="string")
    parser.add_option("--mpicxx",  type="string")
    parser.add_option("--mpifort", type="string")
    parser.add_option("--mpif90",  type="string")
    parser.add_option("--mpif77",  type="string")
    parser.add_option("--mpild",   type="string")
    (opts, args) = parser.parse_args()

    log = dulog.Log(dulog.INFO)
    cfg = Config(log)
    cfg.setup(opts)
    cfg.dump()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
import typing
import datetime
import importlib
sys.path.insert(0, os.path.abspath('.'))
_today = datetime.datetime.now()


# -- Project information -----------------------------------------------------

package = 'mpi4py'


def pkg_version():
    import re
    here = os.path.dirname(__file__)
    pardir = [os.path.pardir] * 3
    topdir = os.path.join(here, *pardir)
    srcdir = os.path.join(topdir, 'src')
    with open(os.path.join(srcdir, 'mpi4py', '__init__.py')) as f:
        m = re.search(r"__version__\s*=\s*'(.*)'", f.read())
        return m.groups()[0]


project = 'MPI for Python'
author = 'Lisandro Dalcin'
copyright = f'{_today.year}, {author}'

# The full version, including alpha/beta/rc tags
release = pkg_version()
version = release.rsplit('.', 1)[0]


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

default_role = 'any'

nitpicky = False
nitpick_ignore = [
    ('c:func', r'atexit'),
    ('envvar', r'MPICC'),
    ('py:mod', r'__worker__'),
    ('py:mod', r'pickle5'),
]
nitpick_ignore_regex = [
    (r'c:.*', r'MPI_.*'),
    (r'envvar', r'(LD_LIBRARY_)?PATH'),
    (r'envvar', r'(MPICH|OMPI|MPIEXEC)_.*'),
]
# python_use_unqualified_type_names = True

autodoc_typehints = 'description'
autodoc_type_aliases = {}
autodoc_mock_imports = []

autosummary_context = {
    'synopsis': {},
    'autotype': {},
}

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
}

napoleon_preprocess_types = True

try:
    import sphinx_rtd_theme
    if 'sphinx_rtd_theme' not in extensions:
        extensions.append('sphinx_rtd_theme')
except ImportError:
    sphinx_rtd_theme = None


def _setup_numpy_typing():
    try:
        import numpy as np
    except ImportError:
        np = type(sys)('numpy')
        sys.modules[np.__name__] = np
        np.dtype = type('dtype', (), {})

    try:
        import numpy.typing as npt
    except ImportError:
        npt = type(sys)('numpy.typing')
        np.typing = npt
        sys.modules[npt.__name__] = npt
        npt.__all__ = []
        for attr in ['ArrayLike', 'DTypeLike']:
            setattr(npt, attr, typing.Any)
            npt.__all__.append(attr)

    autodoc_type_aliases.update({
        'dtype': 'numpy.dtype',
        'ArrayLike': 'numpy.typing.ArrayLike',
        'DTypeLike': 'numpy.typing.DTypeLike',
    })


def _patch_domain_python():
    from sphinx.domains import python
    try:
        from numpy.typing import __all__ as numpy_types
    except ImportError:
        numpy_types = []
    try:
        from mpi4py.typing import __all__ as mpi4py_types
    except ImportError:
        mpi4py_types = []

    def make_xref(self, rolename, domain, target, *args, **kwargs):
        if target in ('True', 'False'):
            rolename = 'obj'
        elif target in numpy_types:
            rolename = 'data'
        elif target in mpi4py_types:
            rolename = 'data'
        return make_xref_orig(self, rolename, domain, target, *args, *kwargs)

    numpy_types = set(f'numpy.typing.{name}' for name in numpy_types)
    mpi4py_types = set(mpi4py_types)
    make_xref_orig = python.PyXrefMixin.make_xref
    python.PyXrefMixin.make_xref = make_xref


def _patch_autosummary():
    from sphinx.ext import autodoc
    from sphinx.ext import autosummary
    from sphinx.ext.autosummary import generate

    class ExceptionDocumenter(autodoc.ExceptionDocumenter):
        objtype = 'class'

    def get_documenter(app, obj, parent):
        if isinstance(obj, type) and issubclass(obj, BaseException):
            caller = sys._getframe().f_back.f_code.co_name
            if caller == 'generate_autosummary_content':
                if obj.__module__ == 'mpi4py.MPI':
                    if obj.__name__ == 'Exception':
                        return ExceptionDocumenter
        return autosummary.get_documenter(app, obj, parent)

    generate.get_documenter = get_documenter


def setup(app):
    _setup_numpy_typing()
    _patch_domain_python()
    _patch_autosummary()
    try:
        from mpi4py import MPI
    except ImportError:
        autodoc_mock_imports.append('mpi4py')
        return

    sys_dwb = sys.dont_write_bytecode
    sys.dont_write_bytecode = True
    import apidoc
    sys.dont_write_bytecode = sys_dwb

    name = MPI.__name__
    here = os.path.abspath(os.path.dirname(__file__))
    outdir = os.path.join(here, apidoc.OUTDIR)
    source = os.path.join(outdir, f'{name}.py')
    getmtime = os.path.getmtime
    generate = (
        not os.path.exists(source)
        or getmtime(source) < getmtime(MPI.__file__)
        or getmtime(source) < getmtime(apidoc.__file__)
    )
    if generate:
        apidoc.generate(source)
    module = apidoc.load_module(source)
    apidoc.replace_module(module)

    for name in dir(module):
        attr = getattr(module, name)
        if isinstance(attr, type):
            if attr.__module__ == module.__name__:
                autodoc_type_aliases[name] = name

    synopsis = autosummary_context['synopsis']
    synopsis[module.__name__] = module.__doc__.strip()
    autotype = autosummary_context['autotype']
    autotype[module.Exception.__name__] = 'exception'


    modules = [
        'mpi4py',
        'mpi4py.run',
        'mpi4py.util.dtlib',
        'mpi4py.util.pkl5',
    ]
    typing_overload = typing.overload
    typing.overload = lambda arg: arg
    for name in modules:
        mod = importlib.import_module(name)
        ann = apidoc.load_module(f'{mod.__file__}i', name)
        apidoc.annotate(mod, ann)
    typing.overload = typing_overload


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = (
    'sphinx_rtd_theme' if 'sphinx_rtd_theme' in extensions else 'default'
)

if html_theme == 'default':
    html_copy_source = False

if html_theme == 'sphinx_rtd_theme':
    html_theme_options = {
        'analytics_id': 'UA-48837848-1',
    }

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

if html_theme == 'sphinx_rtd_theme':
    html_css_files = [
        'css/custom.css',
    ]


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = f'{package}-man'


# -- Options for LaTeX output ---------------------------------------------

# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ('index', f'{package}.tex', project, author, 'howto'),
]

latex_elements = {
    'papersize': 'a4',
}


# -- Options for manual page output ---------------------------------------

# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', package, project, [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index', package, project, author,
     package, f'{project}.', 'Miscellaneous'),
]


# -- Options for Epub output ----------------------------------------------

# Output file base name for ePub builder.
epub_basename = package



cdef class Status:

    """
    Status object
    """

    def __cinit__(self, Status status: Optional[Status] = None):
        cdef MPI_Status *s = &self.ob_mpi
        CHKERR( PyMPI_Status_set_source (s, MPI_ANY_SOURCE ) )
        CHKERR( PyMPI_Status_set_tag    (s, MPI_ANY_TAG    ) )
        CHKERR( PyMPI_Status_set_error  (s, MPI_SUCCESS    ) )
        if status is None: return
        self.ob_mpi = status.ob_mpi

    def __richcmp__(self, other, int op):
        if not isinstance(other, Status): return NotImplemented
        cdef Status s = <Status>self, o = <Status>other
        cdef int ne = memcmp(&s.ob_mpi, &o.ob_mpi, sizeof(MPI_Status))
        if   op == Py_EQ: return (ne == 0)
        elif op == Py_NE: return (ne != 0)
        cdef mod = type(self).__module__
        cdef cls = type(self).__name__
        raise TypeError("unorderable type: '%s.%s'" % (mod, cls))

    def Get_source(self) -> int:
        """
        Get message source
        """
        cdef int source = MPI_ANY_SOURCE
        CHKERR( PyMPI_Status_get_source(&self.ob_mpi, &source) )
        return source

    def Set_source(self, int source: int) -> None:
        """
        Set message source
        """
        CHKERR( PyMPI_Status_set_source(&self.ob_mpi, source) )

    property source:
        """source"""
        def __get__(self) -> int:
            return self.Get_source()
        def __set__(self, value: int):
            self.Set_source(value)

    def Get_tag(self) -> int:
        """
        Get message tag
        """
        cdef int tag = MPI_ANY_TAG
        CHKERR( PyMPI_Status_get_tag(&self.ob_mpi, &tag) )
        return tag

    def Set_tag(self, int tag: int) -> None:
        """
        Set message tag
        """
        CHKERR( PyMPI_Status_set_tag(&self.ob_mpi, tag) )

    property tag:
        """tag"""
        def __get__(self) -> int:
            return self.Get_tag()
        def __set__(self, value: int):
            self.Set_tag(value)

    def Get_error(self) -> int:
        """
        Get message error
        """
        cdef int error = MPI_SUCCESS
        CHKERR( PyMPI_Status_get_error(&self.ob_mpi, &error) )
        return error

    def Set_error(self, int error: int) -> None:
        """
        Set message error
        """
        CHKERR( PyMPI_Status_set_error(&self.ob_mpi, error) )

    property error:
        """error"""
        def __get__(self) -> int:
            return self.Get_error()
        def __set__(self, value: int):
            self.Set_error(value)

    def Get_count(self, Datatype datatype: Datatype = BYTE) -> int:
        """
        Get the number of *top level* elements
        """
        cdef MPI_Datatype dtype = datatype.ob_mpi
        cdef int count = MPI_UNDEFINED
        CHKERR( MPI_Get_count(&self.ob_mpi, dtype, &count) )
        return count

    property count:
        """byte count"""
        def __get__(self) -> int:
            return self.Get_count(__BYTE__)

    def Get_elements(self, Datatype datatype: Datatype) -> int:
        """
        Get the number of basic elements in a datatype
        """
        cdef MPI_Datatype dtype = datatype.ob_mpi
        cdef MPI_Count elements = MPI_UNDEFINED
        CHKERR( MPI_Get_elements_x(&self.ob_mpi, dtype, &elements) )
        return elements

    def Set_elements(
        self,
        Datatype datatype: Datatype,
        Count count: int,
    ) -> None:
        """
        Set the number of elements in a status

        .. note:: This should be only used when implementing
           query callback functions for generalized requests
        """
        cdef MPI_Datatype dtype = datatype.ob_mpi
        CHKERR( MPI_Status_set_elements_x(&self.ob_mpi, dtype, count) )

    def Is_cancelled(self) -> bool:
        """
        Test to see if a request was cancelled
        """
        cdef int flag = 0
        CHKERR( MPI_Test_cancelled(&self.ob_mpi, &flag) )
        return <bint>flag

    def Set_cancelled(self, bint flag: bool) -> None:
        """
        Set the cancelled state associated with a status

        .. note:: This should be only used when implementing
           query callback functions for generalized requests
        """
        CHKERR( MPI_Status_set_cancelled(&self.ob_mpi, flag) )

    property cancelled:
        """
        cancelled state
        """
        def __get__(self) -> bool:
            return self.Is_cancelled()
        def __set__(self, value: int):
            self.Set_cancelled(value)

    # Fortran Handle
    # --------------

    def py2f(self) -> List[int]:
        """
        """
        cdef Status status = <Status> self
        cdef Py_ssize_t i = 0
        cdef Py_ssize_t n = <int>(sizeof(MPI_Status)/sizeof(int))
        cdef MPI_Status *c_status = &status.ob_mpi
        cdef MPI_Fint *f_status = NULL
        cdef tmp = allocate(n+1, sizeof(MPI_Fint), &f_status)
        CHKERR( MPI_Status_c2f(c_status, f_status) )
        return [f_status[i] for i from 0 <= i < n]

    @classmethod
    def f2py(cls, arg: List[int]) -> Status:
        """
        """
        cdef Status status = Status.__new__(Status)
        cdef Py_ssize_t i = 0
        cdef Py_ssize_t n = <int>(sizeof(MPI_Status)/sizeof(int))
        cdef MPI_Status *c_status = &status.ob_mpi
        cdef MPI_Fint *f_status = NULL
        cdef tmp = allocate(n+1, sizeof(MPI_Fint), &f_status)
        for i from 0 <= i < n: f_status[i] = arg[i]
        CHKERR( MPI_Status_f2c(f_status, c_status) )
        return status



## mpiexec -n 2 python ex-2.08.py

# An exchange of messages

# --------------------------------------------------------------------

from mpi4py import MPI
import array

if MPI.COMM_WORLD.Get_size() < 2:
    raise SystemExit

# --------------------------------------------------------------------

sendbuf = array.array('d', [0]*10)
recvbuf = array.array('d', [0]*10)
tag = 0
status = MPI.Status()

myrank = MPI.COMM_WORLD.Get_rank()

if myrank == 0:
    sendbuf[:] = array.array('d', range(len(sendbuf)))
    MPI.COMM_WORLD.Send([sendbuf, MPI.DOUBLE], 1, tag)
    MPI.COMM_WORLD.Recv([recvbuf, MPI.DOUBLE], 1, tag, status)
elif myrank == 1:
    MPI.COMM_WORLD.Recv([recvbuf, MPI.DOUBLE], 0, tag, status)
    sendbuf[:] = recvbuf
    MPI.COMM_WORLD.Send([sendbuf, MPI.DOUBLE], 0, tag)

# --------------------------------------------------------------------

if myrank == 0:
    assert status.source == 1
    assert status.tag == tag
    assert status.error == MPI.SUCCESS
    assert status.Get_count(MPI.DOUBLE) == len(recvbuf)
    assert sendbuf == recvbuf
elif myrank == 1:
    assert status.source == 0
    assert status.tag == tag
    assert status.error == MPI.SUCCESS
    assert status.Get_count(MPI.DOUBLE) == len(recvbuf)
    assert sendbuf == recvbuf

# --------------------------------------------------------------------



import os
import cffi

ffi = cffi.FFI()
with open("libmpi.c.in") as f:
    ffi.set_source("libmpi", f.read())
with open("libmpi.h") as f:
    ffi.cdef(f.read())

class mpicompiler(object):

    from cffi import ffiplatform

    def __init__(self, cc, ld=None):
        self.cc = cc
        self.ld = ld if ld else cc
        self.ffi_compile = self.ffiplatform.compile

    def __enter__(self):
        self.ffiplatform.compile = self.compile

    def __exit__(self, *args):
        self.ffiplatform.compile = self.ffi_compile

    def configure(self, compiler):
        from distutils.util import split_quoted
        from distutils.spawn import find_executable
        def fix_command(command, cmd):
            if not cmd: return
            cmd = split_quoted(cmd)
            exe = find_executable(cmd[0])
            if not exe: return
            command[0] = exe
            command += cmd[1:]
        fix_command(compiler.compiler_so, self.cc)
        fix_command(compiler.linker_so, self.ld)

    def compile(self, *args, **kargs):
        from distutils.command import build_ext
        customize_compiler_orig = build_ext.customize_compiler
        def customize_compiler(compiler):
            customize_compiler_orig(compiler)
            self.configure(compiler)
        build_ext.customize_compiler = customize_compiler
        try:
            return self.ffi_compile(*args, **kargs)
        finally:
            build_ext.customize_compiler = customize_compiler_orig

if __name__ == '__main__':
    cc = os.environ.get('MPICC', 'mpicc')
    ld = os.environ.get('MPILD')
    with mpicompiler(cc, ld):
        ffi.compile()



from mpi4py import MPI
from array import array as _array
import struct as _struct

# --------------------------------------------------------------------

class Counter(object):

    def __init__(self, comm):
        rank = comm.Get_rank()
        itemsize = MPI.INT.Get_size()
        if rank == 0:
            n = 1
        else:
            n = 0
        self.win = MPI.Win.Allocate(n*itemsize, itemsize, 
                                    MPI.INFO_NULL, comm)
        if rank == 0:
            mem = self.win.tomemory()
            mem[:] = _struct.pack('i', 0)

    def free(self):
        self.win.Free()

    def next(self, increment=1):
        incr = _array('i', [increment])
        nval = _array('i', [0])
        self.win.Lock(0)
        self.win.Get_accumulate([incr, 1, MPI.INT], 
                                [nval, 1, MPI.INT],
                                0, op=MPI.SUM)
        self.win.Unlock(0)
        return nval[0]

# -----------------------------------------------------------------------------

class Mutex(object):

    def __init__(self, comm):
        self.counter = Counter(comm)

    def __enter__(self):
        self.lock()
        return self

    def __exit__(self, *exc):
        self.unlock()
        return None

    def free(self):
        self.counter.free()

    def lock(self):
        value = self.counter.next(+1)
        while value != 0:
            value = self.counter.next(-1)
            value = self.counter.next(+1)

    def unlock(self):
        self.counter.next(-1)

# -----------------------------------------------------------------------------

def test_counter():
    vals = []
    counter = Counter(MPI.COMM_WORLD)
    for i in range(5):
        c = counter.next()
        vals.append(c)
    counter.free()

    vals = MPI.COMM_WORLD.allreduce(vals)
    assert sorted(vals) == list(range(len(vals)))

def test_mutex():
    mutex = Mutex(MPI.COMM_WORLD)
    mutex.lock()
    mutex.unlock()
    mutex.free()

if __name__ == '__main__':
    test_counter()
    test_mutex()

# -----------------------------------------------------------------------------



from libmpi import ffi, lib

def ring(comm, count=1, loop=1, skip=0):

    size_p = ffi.new('int*')
    rank_p = ffi.new('int*')
    lib.MPI_Comm_size(comm, size_p)
    lib.MPI_Comm_rank(comm, rank_p)
    size = size_p[0]
    rank = rank_p[0]

    source  = (rank - 1) % size
    dest = (rank + 1) % size
    sbuf = ffi.new('unsigned char[]', [42]*count)
    rbuf = ffi.new('unsigned char[]', [ 0]*count)

    iterations = list(range((loop+skip)))

    if size == 1:
        for i in iterations:
            if i == skip:
                tic = lib.MPI_Wtime()
            lib.MPI_Sendrecv(sbuf, count, lib.MPI_BYTE, dest,   0,
                             rbuf, count, lib.MPI_BYTE, source, 0,
                             comm, lib.MPI_STATUS_IGNORE)
    else:
        if rank == 0:
            for i in iterations:
                if i == skip:
                    tic = lib.MPI_Wtime()
                lib.MPI_Send(sbuf, count, lib.MPI_BYTE, dest,   0, comm)
                lib.MPI_Recv(rbuf, count, lib.MPI_BYTE, source, 0, comm, lib.MPI_STATUS_IGNORE)
        else:
            sbuf = rbuf
            for i in iterations:
                if i == skip:
                    tic = lib.MPI_Wtime()
                lib.MPI_Recv(rbuf, count, lib.MPI_BYTE, source, 0, comm, lib.MPI_STATUS_IGNORE)
                lib.MPI_Send(sbuf, count, lib.MPI_BYTE, dest,   0, comm)
    toc = lib.MPI_Wtime()
    if rank == 0 and ffi.string(sbuf) != ffi.string(rbuf):
        import warnings, traceback
        try:
            warnings.warn("received message does not match!")
        except UserWarning:
            traceback.print_exc()
            lib.MPI_Abort(comm, 2)
    return toc - tic

def ringtest(comm):

    size = ( 1 )
    loop = ( 1 )
    skip = ( 0 )

    lib.MPI_Barrier(comm)
    elapsed = ring(comm, size, loop, skip)

    size_p = ffi.new('int*')
    rank_p = ffi.new('int*')
    lib.MPI_Comm_size(comm, size_p)
    lib.MPI_Comm_rank(comm, rank_p)
    comm_size = size_p[0]
    comm_rank = rank_p[0]

    if comm_rank == 0:
        print ("time for %d loops = %g seconds (%d processes, %d bytes)"
               % (loop, elapsed, comm_size, size))

def main():
    lib.MPI_Init(ffi.NULL, ffi.NULL)
    ringtest(lib.MPI_COMM_WORLD)
    lib.MPI_Finalize()

if __name__ == '__main__':
    main()



from mpi4py import MPI
import mpiunittest as unittest
import arrayimpl

def create_topo_comms(comm):
    size = comm.Get_size()
    rank = comm.Get_rank()
    # Cartesian
    n = int(size**1/2.0)
    m = int(size**1/3.0)
    if m*m*m == size:
        dims = [m, m, m]
    elif n*n == size:
        dims = [n, n]
    else:
        dims = [size]
    periods = [True] * len(dims)
    yield comm.Create_cart(dims, periods=periods)
    # Graph
    index, edges = [0], []
    for i in range(size):
        pos = index[-1]
        index.append(pos+2)
        edges.append((i-1)%size)
        edges.append((i+1)%size)
    yield comm.Create_graph(index, edges)
    # Dist Graph
    sources = [(rank-2)%size, (rank-1)%size]
    destinations = [(rank+1)%size, (rank+2)%size]
    yield comm.Create_dist_graph_adjacent(sources, destinations)

def get_neighbors_count(comm):
    topo = comm.Get_topology()
    if topo == MPI.CART:
        ndim = comm.Get_dim()
        return 2*ndim, 2*ndim
    if topo == MPI.GRAPH:
        rank = comm.Get_rank()
        nneighbors = comm.Get_neighbors_count(rank)
        return nneighbors, nneighbors
    if topo == MPI.DIST_GRAPH:
        indeg, outdeg, w = comm.Get_dist_neighbors_count()
        return indeg, outdeg
    return 0, 0

def have_feature():
    cartcomm = MPI.COMM_SELF.Create_cart([1], periods=[0])
    try:
        cartcomm.neighbor_allgather(None)
        return True
    except NotImplementedError:
        return False
    finally:
        cartcomm.Free()

@unittest.skipIf(not have_feature(), 'mpi-neighbor')
class BaseTestCCONghBuf(object):

    COMM = MPI.COMM_NULL

    def testNeighborAllgather(self):
        for comm in create_topo_comms(self.COMM):
            rsize, ssize = get_neighbors_count(comm)
            for array, typecode in arrayimpl.subTest(self):
                if unittest.is_mpi_gpu('openmpi', array):
                    # segfault as of OpenMPI 4.1.1; TODO(leofang): why?
                    if array.backend == 'numba':
                        continue
                for v in range(3):
                    sbuf = array( v, typecode, 3)
                    rbuf = array(-1, typecode, (rsize, 3))
                    comm.Neighbor_allgather(sbuf.as_mpi(), rbuf.as_mpi())
                    for value in rbuf.flat:
                        self.assertEqual(value, v)
                    sbuf = array( v, typecode, 3)
                    rbuf = array(-1, typecode, (rsize, 3))
                    comm.Neighbor_allgatherv(sbuf.as_mpi_c(3), rbuf.as_mpi_c(3))
                    for value in rbuf.flat:
                        self.assertEqual(value, v)
                    sbuf = array( v, typecode, 3)
                    rbuf = array(-1, typecode, (rsize, 3))
                    comm.Ineighbor_allgather(sbuf.as_mpi(), rbuf.as_mpi()).Wait()
                    for value in rbuf.flat:
                        self.assertEqual(value, v)
                    sbuf = array( v, typecode, 3)
                    rbuf = array(-1, typecode, (rsize, 3))
                    comm.Ineighbor_allgatherv(sbuf.as_mpi_c(3), rbuf.as_mpi_c(3)).Wait()
                    for value in rbuf.flat:
                        self.assertEqual(value, v)
            comm.Free()

    def testNeighborAlltoall(self):
        for comm in create_topo_comms(self.COMM):
            rsize, ssize = get_neighbors_count(comm)
            for array, typecode in arrayimpl.subTest(self):
                for v in range(3):
                    sbuf = array( v, typecode, (ssize, 3))
                    rbuf = array(-1, typecode, (rsize, 3))
                    comm.Neighbor_alltoall(sbuf.as_mpi(), rbuf.as_mpi_c(3))
                    for value in rbuf.flat:
                        self.assertEqual(value, v)
                    sbuf = array( v, typecode, (ssize, 3))
                    rbuf = array(-1, typecode, (rsize, 3))
                    comm.Neighbor_alltoall(sbuf.as_mpi(), rbuf.as_mpi())
                    for value in rbuf.flat:
                        self.assertEqual(value, v)
                    sbuf = array( v, typecode, (ssize, 3))
                    rbuf = array(-1, typecode, (rsize, 3))
                    comm.Neighbor_alltoallv(sbuf.as_mpi_c(3), rbuf.as_mpi_c(3))
                    for value in rbuf.flat:
                        self.assertEqual(value, v)
                    sbuf = array( v, typecode, (ssize, 3))
                    rbuf = array(-1, typecode, (rsize, 3))
                    comm.Ineighbor_alltoall(sbuf.as_mpi(), rbuf.as_mpi()).Wait()
                    for value in rbuf.flat:
                        self.assertEqual(value, v)
                    sbuf = array( v, typecode, (ssize, 3))
                    rbuf = array(-1, typecode, (rsize, 3))
                    comm.Ineighbor_alltoallv(sbuf.as_mpi_c(3), rbuf.as_mpi_c(3)).Wait()
                    for value in rbuf.flat:
                        self.assertEqual(value, v)
            comm.Free()

    def testNeighborAlltoallw(self):
        size = self.COMM.Get_size()
        for comm in create_topo_comms(self.COMM):
            rsize, ssize = get_neighbors_count(comm)
            for array, typecode in arrayimpl.subTest(self):
                for n in range(1,4):
                    for v in range(3):
                        sbuf = array( v, typecode, (ssize, n))
                        rbuf = array(-1, typecode, (rsize, n))
                        sdt, rdt = sbuf.mpidtype, rbuf.mpidtype
                        sdsp = list(range(0, ssize*n*sdt.extent, n*sdt.extent))
                        rdsp = list(range(0, rsize*n*rdt.extent, n*rdt.extent))
                        smsg = [sbuf.as_raw(), ([n]*ssize, sdsp), [sdt]*ssize]
                        rmsg = (rbuf.as_raw(), ([n]*rsize, rdsp), [rdt]*rsize)
                        try:
                            comm.Neighbor_alltoallw(smsg, rmsg)
                        except NotImplementedError:
                            self.skipTest('mpi-neighbor_alltoallw')
                        for value in rbuf.flat:
                            self.assertEqual(value, v)
                        smsg[0] = array(v+1, typecode, (ssize, n)).as_raw()
                        try:
                            comm.Ineighbor_alltoallw(smsg, rmsg).Wait()
                        except NotImplementedError:
                            self.skipTest('mpi-ineighbor_alltoallw')
                        for value in rbuf.flat:
                            self.assertEqual(value, v+1)
            comm.Free()

    def testNeighborAlltoallwBottom(self):
        size = self.COMM.Get_size()
        for comm in create_topo_comms(self.COMM):
            rsize, ssize = get_neighbors_count(comm)
            for array, typecode in arrayimpl.subTest(self):
                for n in range(1,4):
                    for v in range(3):
                        sbuf = array( v, typecode, (ssize, n))
                        rbuf = array(-1, typecode, (rsize, n))
                        saddr = MPI.Get_address(sbuf.as_raw())
                        raddr = MPI.Get_address(rbuf.as_raw())
                        sdt, rdt = sbuf.mpidtype, rbuf.mpidtype
                        sdsp = list(range(0, ssize*n*sdt.extent, n*sdt.extent))
                        rdsp = list(range(0, rsize*n*rdt.extent, n*rdt.extent))
                        sdsp = [saddr + d for d in sdsp]
                        rdsp = [raddr + d for d in rdsp]
                        smsg = [MPI.BOTTOM, ([n]*ssize, sdsp), [sdt]*ssize]
                        rmsg = (MPI.BOTTOM, ([n]*rsize, rdsp), [rdt]*rsize)
                        try:
                            comm.Neighbor_alltoallw(smsg, rmsg)
                        except NotImplementedError:
                            self.skipTest('mpi-neighbor_alltoallw')
                        for value in rbuf.flat:
                            self.assertEqual(value, v)
            comm.Free()


class TestCCONghBufSelf(BaseTestCCONghBuf, unittest.TestCase):
    COMM = MPI.COMM_SELF

class TestCCONghBufWorld(BaseTestCCONghBuf, unittest.TestCase):
    COMM = MPI.COMM_WORLD

class TestCCONghBufSelfDup(TestCCONghBufSelf):
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()
    def tearDown(self):
        self.COMM.Free()

class TestCCONghBufWorldDup(TestCCONghBufWorld):
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()
    def tearDown(self):
        self.COMM.Free()


name, version = MPI.get_vendor()
if name == 'Open MPI':
    if version < (1,8,4):
        _create_topo_comms = create_topo_comms
        def create_topo_comms(comm):
            for c in _create_topo_comms(comm):
                if c.size * 2 < sum(c.degrees):
                    c.Free(); continue
                yield c


if __name__ == '__main__':
    unittest.main()



#------------------------------------------------------------------------------

cdef extern from "Python.h":
    int    PyUnicode_Check(object)
    object PyUnicode_AsUTF8String(object)
    object PyUnicode_AsASCIIString(object)
    object PyUnicode_FromString(const char[])
    object PyUnicode_FromStringAndSize(const char[],Py_ssize_t)
    object PyBytes_FromString(const char[])
    object PyBytes_FromStringAndSize(const char[],Py_ssize_t)
    int    PyBytes_AsStringAndSize(object,char*[],Py_ssize_t*) except -1

#------------------------------------------------------------------------------

cdef inline object asmpistr(object ob, char *s[]):
    if PyUnicode_Check(ob):
        if PY3: ob = PyUnicode_AsUTF8String(ob);
        else:   ob = PyUnicode_AsASCIIString(ob);
    PyBytes_AsStringAndSize(ob, s, NULL)
    return ob

cdef inline object tompistr(const char s[], int n):
    if PY3: return PyUnicode_FromStringAndSize(s, n)
    else:   return PyBytes_FromStringAndSize(s, n)

cdef inline object mpistr(const char s[]):
    if PY3: return PyUnicode_FromString(s)
    else:   return PyBytes_FromString(s)

cdef inline object pystr(const char s[]):
    if PY3: return PyUnicode_FromString(s)
    else:   return PyBytes_FromString(s)

#------------------------------------------------------------------------------



from mpi4py import MPI
import mpiunittest as unittest
import arrayimpl
import sys, os, tempfile


def subTestIO(case, *args, **kwargs):
    for array, typecode in arrayimpl.subTest(case, *args, **kwargs):
        if unittest.is_mpi_gpu('mvapich2', array): continue
        yield array, typecode


class BaseTestIO(object):

    COMM = MPI.COMM_NULL
    FILE = MPI.FILE_NULL

    prefix = 'mpi4py-'

    def setUp(self):
        comm = self.COMM
        fname = None
        if comm.Get_rank() == 0:
            fd, fname = tempfile.mkstemp(prefix=self.prefix)
            os.close(fd)
        fname = comm.bcast(fname, 0)
        amode  = MPI.MODE_RDWR | MPI.MODE_CREATE
        amode |= MPI.MODE_DELETE_ON_CLOSE
        amode |= MPI.MODE_UNIQUE_OPEN
        info = MPI.INFO_NULL
        try:
            self.FILE = MPI.File.Open(comm, fname, amode, info)
        except Exception:
            if comm.Get_rank() == 0:
                os.remove(fname)
            raise

    def tearDown(self):
        if self.FILE:
            self.FILE.Close()
        self.COMM.Barrier()

    # non-collective

    def testReadWriteAt(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(42, typecode, count)
            fh.Write_at(count*rank, wbuf.as_raw())
            fh.Sync()
            comm.Barrier()
            fh.Sync()
            rbuf = array(-1, typecode, count+1)
            fh.Read_at(count*rank, rbuf.as_mpi_c(count))
            for value in rbuf[:-1]:
                self.assertEqual(value, 42)
            self.assertEqual(rbuf[-1], -1)
            comm.Barrier()

    def testIReadIWriteAt(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(42, typecode, count)
            fh.Iwrite_at(count*rank, wbuf.as_raw()).Wait()
            fh.Sync()
            comm.Barrier()
            fh.Sync()
            rbuf = array(-1, typecode, count+1)
            fh.Iread_at(count*rank, rbuf.as_mpi_c(count)).Wait()
            for value in rbuf[:-1]:
                self.assertEqual(value, 42)
            self.assertEqual(rbuf[-1], -1)
            comm.Barrier()

    def testReadWrite(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(42, typecode, count)
            for r in range(size):
                if r == rank:
                    fh.Seek(0, MPI.SEEK_SET)
                    fh.Write(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                for n in range(0, len(wbuf)):
                    rbuf = array(-1, typecode, n+1)
                    fh.Seek(0, MPI.SEEK_SET)
                    fh.Read(rbuf.as_mpi_c(n))
                    for value in rbuf[:-1]:
                        self.assertEqual(value, 42)
                    self.assertEqual(rbuf[-1], -1)
                comm.Barrier()

    def testIReadIWrite(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(42, typecode, count)
            for r in range(size):
                if r == rank:
                    fh.Seek(0, MPI.SEEK_SET)
                    fh.Iwrite(wbuf.as_raw()).Wait()
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                for n in range(0, len(wbuf)):
                    rbuf = array(-1, typecode, n+1)
                    fh.Seek(0, MPI.SEEK_SET)
                    fh.Iread(rbuf.as_mpi_c(n)).Wait()
                    for value in rbuf[:-1]:
                        self.assertEqual(value, 42)
                    self.assertEqual(rbuf[-1], -1)
                comm.Barrier()

    def testReadWriteShared(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(rank%42, typecode, count)
            fh.Seek_shared(0, MPI.SEEK_SET)
            fh.Write_shared(wbuf.as_raw())
            fh.Sync()
            comm.Barrier()
            fh.Sync()
            rbuf = array(-1, typecode, count+1)
            fh.Seek_shared(0, MPI.SEEK_SET)
            fh.Read_shared(rbuf.as_mpi_c(count))
            for value in rbuf[:-1]:
                self.assertTrue(0<=value<42)
                self.assertEqual(value, rbuf[0])
            self.assertEqual(rbuf[-1], -1)
            comm.Barrier()

    def testIReadIWriteShared(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(rank%42, typecode, count)
            fh.Seek_shared(0, MPI.SEEK_SET)
            fh.Iwrite_shared(wbuf.as_raw()).Wait()
            fh.Sync()
            comm.Barrier()
            fh.Sync()
            rbuf = array(-1, typecode, count+1)
            fh.Seek_shared(0, MPI.SEEK_SET)
            fh.Iread_shared(rbuf.as_mpi_c(count)).Wait()
            for value in rbuf[:-1]:
                self.assertTrue(0<=value<42)
                self.assertEqual(value, rbuf[0])
            self.assertEqual(rbuf[-1], -1)
            comm.Barrier()

    # collective

    def testReadWriteAtAll(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(42, typecode, count)
            fh.Write_at_all(count*rank, wbuf.as_raw())
            fh.Sync()
            comm.Barrier()
            fh.Sync()
            rbuf = array(-1, typecode, count+1)
            fh.Read_at_all(count*rank, rbuf.as_mpi_c(count))
            for value in rbuf[:-1]:
                self.assertEqual(value, 42)
            self.assertEqual(rbuf[-1], -1)
            comm.Barrier()

    @unittest.skipMPI('SpectrumMPI')
    def testIReadIWriteAtAll(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        try: # MPI 3.1
            for array, typecode in subTestIO(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(42, typecode, count)
                fh.Iwrite_at_all(count*rank, wbuf.as_raw()).Wait()
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count+1)
                fh.Iread_at_all(count*rank, rbuf.as_mpi_c(count)).Wait()
                for value in rbuf[:-1]:
                    self.assertEqual(value, 42)
                self.assertEqual(rbuf[-1], -1)
                comm.Barrier()
        except NotImplementedError:
            if MPI.Get_version() >= (3, 1): raise
            self.skipTest('mpi-iwrite_at_all')

    def testReadWriteAtAllBeginEnd(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(42, typecode, count)
            fh.Write_at_all_begin(count*rank, wbuf.as_raw())
            fh.Write_at_all_end(wbuf.as_raw())
            fh.Sync()
            comm.Barrier()
            fh.Sync()
            rbuf = array(-1, typecode, count+1)
            fh.Read_at_all_begin(count*rank, rbuf.as_mpi_c(count))
            fh.Read_at_all_end(rbuf.as_raw())
            for value in rbuf[:-1]:
                self.assertEqual(value, 42)
            self.assertEqual(rbuf[-1], -1)
            comm.Barrier()

    def testReadWriteAll(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(42, typecode, count)
            fh.Seek(count*rank, MPI.SEEK_SET)
            fh.Write_all(wbuf.as_raw())
            fh.Sync()
            comm.Barrier()
            fh.Sync()
            rbuf = array(-1, typecode, count+1)
            fh.Seek(count*rank, MPI.SEEK_SET)
            fh.Read_all(rbuf.as_mpi_c(count))
            for value in rbuf[:-1]:
                self.assertEqual(value, 42)
            self.assertEqual(rbuf[-1], -1)
            comm.Barrier()

    @unittest.skipMPI('SpectrumMPI')
    def testIReadIWriteAll(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        try: # MPI 3.1
            for array, typecode in subTestIO(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(42, typecode, count)
                fh.Seek(count*rank, MPI.SEEK_SET)
                fh.Iwrite_all(wbuf.as_raw()).Wait()
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count+1)
                fh.Seek(count*rank, MPI.SEEK_SET)
                fh.Iread_all(rbuf.as_mpi_c(count)).Wait()
                for value in rbuf[:-1]:
                    self.assertEqual(value, 42)
                self.assertEqual(rbuf[-1], -1)
                comm.Barrier()
        except NotImplementedError:
            if MPI.Get_version() >= (3, 1): raise
            self.skipTest('mpi-iwrite_all')

    def testReadWriteAllBeginEnd(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(42, typecode, count)
            fh.Seek(count*rank, MPI.SEEK_SET)
            fh.Write_all_begin(wbuf.as_raw())
            fh.Write_all_end(wbuf.as_raw())
            fh.Sync()
            comm.Barrier()
            fh.Sync()
            rbuf = array(-1, typecode, count+1)
            fh.Seek(count*rank, MPI.SEEK_SET)
            fh.Read_all_begin(rbuf.as_mpi_c(count))
            fh.Read_all_end(rbuf.as_raw())
            for value in rbuf[:-1]:
                self.assertEqual(value, 42)
            self.assertEqual(rbuf[-1], -1)
            comm.Barrier()

    def testReadWriteOrdered(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(rank%42, typecode, count)
            fh.Seek_shared(0, MPI.SEEK_SET)
            fh.Write_ordered(wbuf.as_raw())
            fh.Sync()
            comm.Barrier()
            fh.Sync()
            rbuf = array(-1, typecode, count+1)
            fh.Seek_shared(0, MPI.SEEK_SET)
            fh.Read_ordered(rbuf.as_mpi_c(count))
            for value in rbuf[:-1]:
                self.assertEqual(value, rank%42)
            self.assertEqual(rbuf[-1], -1)
            comm.Barrier()

    def testReadWriteOrderedBeginEnd(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in subTestIO(self):
            etype = array.TypeMap[typecode]
            fh.Set_size(0)
            fh.Set_view(0, etype)
            count = 13
            wbuf = array(rank%42, typecode, count)
            fh.Seek_shared(0, MPI.SEEK_SET)
            fh.Write_ordered_begin(wbuf.as_raw())
            fh.Write_ordered_end(wbuf.as_raw())
            fh.Sync()
            comm.Barrier()
            fh.Sync()
            rbuf = array(-1, typecode, count+1)
            fh.Seek_shared(0, MPI.SEEK_SET)
            fh.Read_ordered_begin(rbuf.as_mpi_c(count))
            fh.Read_ordered_end(rbuf.as_raw())
            for value in rbuf[:-1]:
                self.assertEqual(value, rank%42)
            self.assertEqual(rbuf[-1], -1)
            comm.Barrier()

@unittest.skipMPI('MPICH1')
@unittest.skipMPI('LAM/MPI')
class TestIOSelf(BaseTestIO, unittest.TestCase):
    COMM = MPI.COMM_SELF
    prefix = BaseTestIO.prefix + ('%d-' % MPI.COMM_WORLD.Get_rank())

@unittest.skipMPI('openmpi(<2.2.0)')
@unittest.skipMPI('msmpi')
@unittest.skipMPI('MPICH2')
@unittest.skipMPI('MPICH1')
@unittest.skipMPI('LAM/MPI')
class TestIOWorld(BaseTestIO, unittest.TestCase):
    COMM = MPI.COMM_WORLD


def have_feature():
    case = BaseTestIO()
    case.COMM = TestIOSelf.COMM
    case.prefix = TestIOSelf.prefix
    case.setUp()
    case.tearDown()
try:
    have_feature()
except NotImplementedError:
    unittest.disable(BaseTestIO, 'mpi-io')


if __name__ == '__main__':
    unittest.main()



from mpi4py import MPI
import mpiunittest as unittest

class TestCommNull(unittest.TestCase):

    def testContructor(self):
        comm = MPI.Comm()
        self.assertEqual(comm, MPI.COMM_NULL)
        self.assertFalse(comm is MPI.COMM_NULL)
        def construct(): MPI.Comm((1,2,3))
        self.assertRaises(TypeError, construct)

    def testContructorIntra(self):
        comm_null = MPI.Intracomm()
        self.assertFalse(comm_null is MPI.COMM_NULL)
        self.assertEqual(comm_null, MPI.COMM_NULL)

    def testContructorInter(self):
        comm_null = MPI.Intercomm()
        self.assertFalse(comm_null is MPI.COMM_NULL)
        self.assertEqual(comm_null, MPI.COMM_NULL)

class BaseTestComm(object):

    def testContructor(self):
        comm = MPI.Comm(self.COMM)
        self.assertEqual(comm, self.COMM)
        self.assertFalse(comm is self.COMM)

    def testPyProps(self):
        comm = self.COMM
        self.assertEqual(comm.Get_size(), comm.size)
        self.assertEqual(comm.Get_rank(), comm.rank)
        self.assertEqual(comm.Is_intra(), comm.is_intra)
        self.assertEqual(comm.Is_inter(), comm.is_inter)
        self.assertEqual(comm.Get_topology(), comm.topology)

    def testGroup(self):
        comm  = self.COMM
        group = self.COMM.Get_group()
        self.assertEqual(comm.Get_size(), group.Get_size())
        self.assertEqual(comm.Get_rank(), group.Get_rank())
        group.Free()
        self.assertEqual(group, MPI.GROUP_NULL)

    def testCloneFree(self):
        comm = self.COMM.Clone()
        comm.Free()
        self.assertEqual(comm, MPI.COMM_NULL)

    def testCompare(self):
        results = (MPI.IDENT, MPI.CONGRUENT, MPI.SIMILAR, MPI.UNEQUAL)
        ccmp = MPI.Comm.Compare(self.COMM, MPI.COMM_WORLD)
        self.assertTrue(ccmp in results)
        ccmp = MPI.Comm.Compare(self.COMM, self.COMM)
        self.assertEqual(ccmp, MPI.IDENT)
        comm = self.COMM.Dup()
        ccmp = MPI.Comm.Compare(self.COMM, comm)
        comm.Free()
        self.assertEqual(ccmp, MPI.CONGRUENT)

    def testIsInter(self):
        is_inter = self.COMM.Is_inter()
        self.assertTrue(type(is_inter) is bool)

    def testGetSetName(self):
        try:
            name = self.COMM.Get_name()
            self.COMM.Set_name('comm')
            self.assertEqual(self.COMM.Get_name(), 'comm')
            self.COMM.Set_name(name)
            self.assertEqual(self.COMM.Get_name(), name)
        except NotImplementedError:
            self.skipTest('mpi-comm-name')

    def testGetParent(self):
        try:
            parent = MPI.Comm.Get_parent()
        except NotImplementedError:
            self.skipTest('mpi-comm-get_parent')

    def testDupWithInfo(self):
        info = None
        self.COMM.Dup(info).Free()
        info = MPI.INFO_NULL
        self.COMM.Dup(info).Free()
        self.COMM.Dup_with_info(info).Free()
        info = MPI.Info.Create()
        self.COMM.Dup(info).Free()
        self.COMM.Dup_with_info(info).Free()
        info.Free()

    @unittest.skipMPI('mpich(<=3.1.0)', MPI.Query_thread() > MPI.THREAD_SINGLE)
    def testIDup(self):
        try:
            comm, request = self.COMM.Idup()
        except NotImplementedError:
            self.skipTest('mpi-comm-idup')
        request.Wait()
        ccmp = MPI.Comm.Compare(self.COMM, comm)
        comm.Free()
        self.assertEqual(ccmp, MPI.CONGRUENT)

    def testGetSetInfo(self):
        #info = MPI.INFO_NULL
        #self.COMM.Set_info(info)
        info = MPI.Info.Create()
        self.COMM.Set_info(info)
        info.Free()
        info = self.COMM.Get_info()
        self.COMM.Set_info(info)
        info.Free()

    def testCreate(self):
        group = self.COMM.Get_group()
        comm = self.COMM.Create(group)
        ccmp = MPI.Comm.Compare(self.COMM, comm)
        self.assertEqual(ccmp, MPI.CONGRUENT)
        comm.Free()
        group.Free()

    @unittest.skipMPI('openmpi(<=1.8.1)')
    def testCreateGroup(self):
        group = self.COMM.Get_group()
        try:
            try:
                comm = self.COMM.Create_group(group)
                ccmp = MPI.Comm.Compare(self.COMM, comm)
                self.assertEqual(ccmp, MPI.CONGRUENT)
                comm.Free()
            finally:
                group.Free()
        except NotImplementedError:
            self.skipTest('mpi-comm-create_group')


    @unittest.skipMPI('openmpi(==2.0.0)')
    def testSplitType(self):
        try:
            MPI.COMM_SELF.Split_type(MPI.COMM_TYPE_SHARED).Free()
        except NotImplementedError:
            self.skipTest('mpi-comm-split_type')
        #comm = self.COMM.Split_type(MPI.UNDEFINED)
        #self.assertEqual(comm, MPI.COMM_NULL)
        comm = self.COMM.Split_type(MPI.COMM_TYPE_SHARED)
        self.assertNotEqual(comm, MPI.COMM_NULL)
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        if size == 1:
            self.assertEqual(comm.size, 1)
            self.assertEqual(comm.rank, 0)
        comm.Free()
        for root in range(size):
            if rank == root:
                split_type = MPI.COMM_TYPE_SHARED
            else:
                split_type = MPI.UNDEFINED
            comm = self.COMM.Split_type(split_type)
            if rank == root:
                self.assertNotEqual(comm, MPI.COMM_NULL)
                self.assertEqual(comm.size, 1)
                self.assertEqual(comm.rank, 0)
                comm.Free()
            else:
                self.assertEqual(comm, MPI.COMM_NULL)


class TestCommSelf(BaseTestComm, unittest.TestCase):
    def setUp(self):
        self.COMM = MPI.COMM_SELF
    def testSize(self):
        size = self.COMM.Get_size()
        self.assertEqual(size, 1)
    def testRank(self):
        rank = self.COMM.Get_rank()
        self.assertEqual(rank, 0)

class TestCommWorld(BaseTestComm, unittest.TestCase):
    def setUp(self):
        self.COMM = MPI.COMM_WORLD
    def testSize(self):
        size = self.COMM.Get_size()
        self.assertTrue(size >= 1)
    def testRank(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        self.assertTrue(rank >= 0 and rank < size)

class TestCommSelfDup(TestCommSelf):
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()
    def tearDown(self):
        self.COMM.Free()

@unittest.skipMPI('openmpi(<1.4.0)', MPI.Query_thread() > MPI.THREAD_SINGLE)
class TestCommWorldDup(TestCommWorld):
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()
    def tearDown(self):
        self.COMM.Free()


if __name__ == '__main__':
    unittest.main()



from mpi4py import MPI
import mpiunittest as unittest

def appnum():
    if MPI.APPNUM == MPI.KEYVAL_INVALID: return None
    return MPI.COMM_WORLD.Get_attr(MPI.APPNUM)

class TestEnviron(unittest.TestCase):

    def testIsInitialized(self):
        flag = MPI.Is_initialized()
        self.assertTrue(type(flag) is bool)
        self.assertTrue(flag)

    def testIsFinalized(self):
        flag = MPI.Is_finalized()
        self.assertTrue(type(flag) is bool)
        self.assertFalse(flag)

    def testGetVersion(self):
        version = MPI.Get_version()
        self.assertEqual(len(version), 2)
        major, minor = version
        self.assertTrue(type(major) is int)
        self.assertTrue(major >= 1)
        self.assertTrue(type(minor) is int)
        self.assertTrue(minor >= 0)

    def testGetLibraryVersion(self):
        version = MPI.Get_library_version()
        self.assertTrue(isinstance(version, str))
        self.assertTrue(len(version) > 0)

    def testGetProcessorName(self):
        procname = MPI.Get_processor_name()
        self.assertTrue(isinstance(procname, str))

    def testWTime(self):
        time1 = MPI.Wtime()
        self.assertTrue(type(time1) is float)
        time2 = MPI.Wtime()
        self.assertTrue(type(time2) is float)
        self.assertTrue(time2 >= time1)

    def testWTick(self):
        tick = MPI.Wtick()
        self.assertTrue(type(tick) is float)
        self.assertTrue(tick > 0.0)

    def testPControl(self):
        for level in (2, 1, 0):
            MPI.Pcontrol(level)
        MPI.Pcontrol(1)

class TestWorldAttrs(unittest.TestCase):

    def testWTimeIsGlobal(self):
        wtg = MPI.COMM_WORLD.Get_attr(MPI.WTIME_IS_GLOBAL)
        if wtg is not None:
            self.assertTrue(wtg in (True, False))

    def testWTimeIsGlobal(self):
        wtg = MPI.COMM_WORLD.Get_attr(MPI.WTIME_IS_GLOBAL)
        if wtg is not None:
            self.assertTrue(wtg in (True, False))

    def testHostPorcessor(self):
        size = MPI.COMM_WORLD.Get_size()
        vals = list(range(size)) + [MPI.PROC_NULL]
        hostproc = MPI.COMM_WORLD.Get_attr(MPI.HOST)
        if hostproc is not None:
            self.assertTrue(hostproc in vals)

    def testIOProcessor(self):
        size = MPI.COMM_WORLD.Get_size()
        vals = list(range(size)) + [MPI.UNDEFINED,
                                    MPI.ANY_SOURCE,
                                    MPI.PROC_NULL]
        ioproc = MPI.COMM_WORLD.Get_attr(MPI.IO)
        if ioproc is not None:
            self.assertTrue(ioproc in vals)

    @unittest.skipIf(MPI.APPNUM == MPI.KEYVAL_INVALID, 'mpi-appnum')
    def testAppNum(self):
        appnum = MPI.COMM_WORLD.Get_attr(MPI.APPNUM)
        if appnum is not None:
            self.assertTrue(appnum == MPI.UNDEFINED or appnum >= 0)

    @unittest.skipMPI('MPICH(>1.2.0)', appnum() is None)
    @unittest.skipMPI('MVAPICH2', appnum() is None)
    @unittest.skipMPI('MPICH2', appnum() is None)
    @unittest.skipIf(MPI.UNIVERSE_SIZE == MPI.KEYVAL_INVALID, 'mpi-universe-size')
    def testUniverseSize(self):
        univsz = MPI.COMM_WORLD.Get_attr(MPI.UNIVERSE_SIZE)
        if univsz is not None:
            self.assertTrue(univsz == MPI.UNDEFINED or univsz >= 0)

    @unittest.skipIf(MPI.LASTUSEDCODE == MPI.KEYVAL_INVALID, 'mpi-lastusedcode')
    def testLastUsedCode(self):
        lastuc = MPI.COMM_WORLD.Get_attr(MPI.LASTUSEDCODE)
        self.assertTrue(lastuc >= 0)


if __name__ == '__main__':
    unittest.main()



cdef class Op:

    """
    Operation object
    """

    def __cinit__(self, Op op: Optional[Op] = None):
        self.ob_mpi = MPI_OP_NULL
        if op is None: return
        self.ob_mpi = op.ob_mpi
        self.ob_func = op.ob_func
        self.ob_usrid = op.ob_usrid

    def __dealloc__(self):
        if not (self.flags & PyMPI_OWNED): return
        CHKERR( del_Op(&self.ob_mpi) )
        op_user_del(&self.ob_usrid)

    def __richcmp__(self, other, int op):
        if not isinstance(other, Op): return NotImplemented
        cdef Op s = <Op>self, o = <Op>other
        if   op == Py_EQ: return (s.ob_mpi == o.ob_mpi)
        elif op == Py_NE: return (s.ob_mpi != o.ob_mpi)
        cdef mod = type(self).__module__
        cdef cls = type(self).__name__
        raise TypeError("unorderable type: '%s.%s'" % (mod, cls))

    def __bool__(self) -> bool:
        return self.ob_mpi != MPI_OP_NULL

    def __call__(self, x: Any, y: Any) -> Any:
        if self.ob_func != NULL:
            return self.ob_func(x, y)
        else:
            return op_user_py(self.ob_usrid, x, y, None)

    @classmethod
    def Create(
        cls,
        function: Callable[[Buffer, Buffer, Datatype], None],
        bint commute: bool = False,
    ) -> Op:
        """
        Create a user-defined operation
        """
        cdef Op op = Op.__new__(Op)
        cdef MPI_User_function *cfunction = NULL
        op.ob_usrid = op_user_new(function, &cfunction)
        CHKERR( MPI_Op_create(cfunction, commute, &op.ob_mpi) )
        return op

    def Free(self) -> None:
        """
        Free the operation
        """
        CHKERR( MPI_Op_free(&self.ob_mpi) )
        op_user_del(&self.ob_usrid)
        if   self is __MAX__     : self.ob_mpi =  MPI_MAX
        elif self is __MIN__     : self.ob_mpi =  MPI_MIN
        elif self is __SUM__     : self.ob_mpi =  MPI_SUM
        elif self is __PROD__    : self.ob_mpi =  MPI_PROD
        elif self is __LAND__    : self.ob_mpi =  MPI_LAND
        elif self is __BAND__    : self.ob_mpi =  MPI_BAND
        elif self is __LOR__     : self.ob_mpi =  MPI_LOR
        elif self is __BOR__     : self.ob_mpi =  MPI_BOR
        elif self is __LXOR__    : self.ob_mpi =  MPI_LXOR
        elif self is __BXOR__    : self.ob_mpi =  MPI_BXOR
        elif self is __MAXLOC__  : self.ob_mpi =  MPI_MAXLOC
        elif self is __MINLOC__  : self.ob_mpi =  MPI_MINLOC
        elif self is __REPLACE__ : self.ob_mpi =  MPI_REPLACE
        elif self is __NO_OP__   : self.ob_mpi =  MPI_NO_OP

    # Process-local reduction
    # -----------------------

    def Is_commutative(self) -> bool:
        """
        Query reduction operations for their commutativity
        """
        cdef int flag = 0
        CHKERR( MPI_Op_commutative(self.ob_mpi, &flag) )
        return <bint>flag

    property is_commutative:
        """is commutative"""
        def __get__(self) -> bool:
            return self.Is_commutative()

    def Reduce_local(self, inbuf: BufSpec, inoutbuf: BufSpec) -> None:
        """
        Apply a reduction operator to local data
        """
        # get *in* and *inout* buffers
        cdef _p_msg_cco m = message_cco()
        m.for_cro_send(inbuf, 0)
        m.for_cro_recv(inoutbuf, 0)
        # check counts and datatypes
        if m.scount != m.rcount: raise ValueError(
            "mismatch in inbuf count %d and inoutbuf count %d" %
            (m.scount, m.rcount))
        if (m.stype != m.rtype): raise ValueError(
            "mismatch in inbuf and inoutbuf MPI datatypes")
        # do local reduction
        with nogil: CHKERR( MPI_Reduce_local(
            m.sbuf, m.rbuf, m.rcount, m.rtype, self.ob_mpi) )

    property is_predefined:
        """is a predefined operation"""
        def __get__(self) -> bool:
            cdef MPI_Op op = self.ob_mpi
            return (op == MPI_OP_NULL or
                    op == MPI_MAX or
                    op == MPI_MIN or
                    op == MPI_SUM or
                    op == MPI_PROD or
                    op == MPI_LAND or
                    op == MPI_BAND or
                    op == MPI_LOR or
                    op == MPI_BOR or
                    op == MPI_LXOR or
                    op == MPI_BXOR or
                    op == MPI_MAXLOC or
                    op == MPI_MINLOC or
                    op == MPI_REPLACE or
                    op == MPI_NO_OP)

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Op_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Op:
        """
        """
        cdef Op op = Op.__new__(Op)
        op.ob_mpi = MPI_Op_f2c(arg)
        return op



cdef Op __OP_NULL__ = new_Op( MPI_OP_NULL )
cdef Op __MAX__     = new_Op( MPI_MAX     )
cdef Op __MIN__     = new_Op( MPI_MIN     )
cdef Op __SUM__     = new_Op( MPI_SUM     )
cdef Op __PROD__    = new_Op( MPI_PROD    )
cdef Op __LAND__    = new_Op( MPI_LAND    )
cdef Op __BAND__    = new_Op( MPI_BAND    )
cdef Op __LOR__     = new_Op( MPI_LOR     )
cdef Op __BOR__     = new_Op( MPI_BOR     )
cdef Op __LXOR__    = new_Op( MPI_LXOR    )
cdef Op __BXOR__    = new_Op( MPI_BXOR    )
cdef Op __MAXLOC__  = new_Op( MPI_MAXLOC  )
cdef Op __MINLOC__  = new_Op( MPI_MINLOC  )
cdef Op __REPLACE__ = new_Op( MPI_REPLACE )
cdef Op __NO_OP__   = new_Op( MPI_NO_OP   )


# Predefined operation handles
# ----------------------------

OP_NULL = __OP_NULL__  #: Null
MAX     = __MAX__      #: Maximum
MIN     = __MIN__      #: Minimum
SUM     = __SUM__      #: Sum
PROD    = __PROD__     #: Product
LAND    = __LAND__     #: Logical and
BAND    = __BAND__     #: Bit-wise and
LOR     = __LOR__      #: Logical or
BOR     = __BOR__      #: Bit-wise or
LXOR    = __LXOR__     #: Logical xor
BXOR    = __BXOR__     #: Bit-wise xor
MAXLOC  = __MAXLOC__   #: Maximum and location
MINLOC  = __MINLOC__   #: Minimum and location
REPLACE = __REPLACE__  #: Replace (for RMA)
NO_OP   = __NO_OP__    #: No-op   (for RMA)
